"""
.. module:: likelihood

:Synopsis: Likelihood class and likelihood collection
:Author: Jesus Torrado and Antony Lewis

This module defines the main :class:`Likelihood` class, from which every likelihood
usually inherits, and the :class:`LikelihoodCollection` class, which groups and manages
all the individual likelihoods.

Likelihoods inherit from :class:`~theory.Theory`, adding an additional method
to return the likelihood. As with all theories, likelihoods cache results, and the
function :meth:`LikelihoodInterface.current_logp` is used by :class:`model.Model` to
calculate the total likelihood. The default Likelihood implementation does the actual
calculation of the log likelihood in the `logp` function, which is then called
by :meth:`Likelihood.calculate` to save the result into the current state.

Subclasses typically just provide the `logp` function to define their likelihood result,
and use :meth:`~theory.Theory.get_requirements` to specify which inputs are needed from
other theory codes (or likelihoods). Other methods of the :class:`~theory.Theory` base
class can be used as and when needed.

"""

# Global
from time import sleep
from typing import Mapping, Optional, Union, Dict
from itertools import chain
import numpy as np
import numbers

# Local
from cobaya.typing import LikesDict, LikeDictIn, ParamValuesDict, empty_dict
from cobaya.tools import get_external_function, getfullargspec, str_to_list
from cobaya.log import LoggedError
from cobaya.component import ComponentCollection, get_component_class
from cobaya.theory import Theory


class LikelihoodInterface:
    """
    Interface function for likelihoods. Can descend from a :class:`~theory.Theory` class
    and this to make a likelihood (where the calculate() method stores state['logp'] for
    the current parameters), or likelihoods can directly inherit from :class:`Likelihood`
    instead.

    The current_logp property returns the current state's logp, and does not normally
    need to be changed.
    """

    current_state: Dict

    @property
    def current_logp(self) -> float:
        """
        Gets log likelihood for the current point

        :return:  log likelihood from the current state
        """
        return self.current_state["logp"]


def is_LikelihoodInterface(class_instance):
    """
    Checks for `current_logp` property. `hasattr()` cannot safely be used in this case
    because `self._current_state` has not yet been defined.

    Works for both classes and instances.
    """
    # NB: This is much faster than "<method> in dir(class)"
    cls = class_instance if class_instance.__class__ is type \
        else class_instance.__class__
    return isinstance(getattr(cls, "current_logp", None), property)


class Likelihood(Theory, LikelihoodInterface):
    """Base class for likelihoods. Extends from :class:`LikelihoodInterface` and the
    general :class:`~theory.Theory` class by adding functions to return likelihoods
    functions (logp function for a given point)."""

    type: Optional[Union[list, str]] = []

    def __init__(self, info: LikeDictIn = empty_dict,
                 name: Optional[str] = None,
                 timing: Optional[bool] = None,
                 packages_path: Optional[str] = None,
                 initialize=True, standalone=True):
        self.delay = 0
        super().__init__(info, name=name, timing=timing,
                         packages_path=packages_path, initialize=initialize,
                         standalone=standalone)

    @property
    def theory(self):
        # for backwards compatibility
        return self.provider

    def logp(self, **params_values):
        """
        Computes and returns the log likelihood value.
        Takes as keyword arguments the parameter values.
        To get the derived parameters, pass a `_derived` keyword with an empty dictionary.

        Alternatively you can just implement calculate() and save the log likelihood into
        state['logp']; this may be more convenient if you also need to also calculate
        other quantities.
        """
        return None

    def marginal(self, directions=None, params_values=None):
        """
        (For analytic likelihoods only.)
        Computes the marginal likelihood.
        If nothing is specified, returns the total marginal likelihood.
        If some directions are specified (as a list, tuple or array), returns the marginal
        likelihood pdf over those directions evaluated at the given parameter values.
        """
        raise LoggedError(self.log, "Exact marginal likelihood not defined.")

    def calculate(self, state, want_derived=True, **params_values_dict):
        """
        Calculates the likelihood and any derived parameters or needs.
        Return False is the calculation fails.

        """
        derived: Optional[ParamValuesDict] = {} if want_derived else None
        state["logp"] = -np.inf  # in case of exception
        state["logp"] = self.logp(_derived=derived, **params_values_dict)
        self.log.debug("Computed log-likelihood = %g", state["logp"])
        if derived is not None:
            state["derived"] = derived.copy()

    def wait(self):
        if self.delay:
            self.log.debug("Sleeping for %f seconds.", self.delay)
            sleep(self.delay)


class AbsorbUnusedParamsLikelihood(Likelihood):
    pass


class LikelihoodExternalFunction(Likelihood):
    def __init__(self, info, name, timing=None):
        Theory.__init__(self, info, name=name, timing=timing, standalone=False)
        # Store the external function and assign its arguments
        self.external_function = get_external_function(info["external"], name=name)
        self._self_arg = "_self"
        argspec = getfullargspec(self.external_function)
        self.input_params = str_to_list(self.input_params)
        ignore_args = [self._self_arg]
        if argspec.defaults:
            required_args = argspec.args[:-len(argspec.defaults)]
        else:
            required_args = argspec.args
        self.params = {p: None for p in required_args if p not in ignore_args}
        if self.output_params:
            self.output_params = str_to_list(self.output_params) or []
        # Required quantities from other components
        self._uses_self_arg = self._self_arg in argspec.args
        if info.get("requires") and not self._uses_self_arg:
            raise LoggedError(
                self.log, "If a likelihood has external requirements, declared under %r, "
                          "it needs to accept a keyword argument %r.", "requires",
                self._self_arg)
        self._requirements = info.get("requires") or {}
        self._optional_args = \
            [p for p, val in chain(zip(argspec.args[-len(argspec.defaults):],
                                       argspec.defaults) if argspec.defaults else [],
                                   (argspec.kwonlydefaults or {}).items())
             if p not in ignore_args and
             (isinstance(val, numbers.Number) or val is None)]
        self._args = set(chain(self._optional_args, self.params))
        if argspec.varkw:
            self._args.update(self.input_params)
        self.log.info("Initialized external likelihood.")

    def get_requirements(self):
        return self._requirements

    def get_can_support_params(self):
        return self._optional_args

    def logp(self, **params_values):
        # Remove non-input params (except _derived)
        _derived = params_values.pop("_derived", None)
        for p in list(params_values):
            if p not in self._args:
                params_values.pop(p)
        if self._uses_self_arg:
            params_values[self._self_arg] = self
        try:
            return_value = self.external_function(**params_values)
        except:
            self.log.debug("External function failed at evaluation.")
            raise
        bad_return_msg = "Expected return value `(logp, {derived_params_dict})`."
        if hasattr(return_value, "__len__"):
            logp = return_value[0]
            if self.output_params:
                try:
                    if _derived is not None:
                        _derived.update(return_value[1])
                        params_values["_derived"] = _derived
                except:
                    raise LoggedError(self.log, bad_return_msg)
        elif self.output_params:
            raise LoggedError(self.log, bad_return_msg)
        else:  # no return.__len__ and output_params expected
            logp = return_value
        return logp


class LikelihoodCollection(ComponentCollection):
    """
    A dictionary storing experimental likelihood :class:`Likelihood` instances index
    by their names.
    """

    def __init__(self, info_likelihood: LikesDict, packages_path=None, timing=None,
                 theory=None):
        super().__init__()
        self.set_logger("likelihood")
        self.theory = theory
        # Get the individual likelihood classes
        for name, info in info_likelihood.items():
            if isinstance(name, Theory):
                name, info = name.get_name(), info
            if isinstance(info, Theory):
                self.add_instance(name, info)
            elif isinstance(info, Mapping) and "external" in info:
                external = info["external"]
                if isinstance(external, Theory):
                    self.add_instance(name, external)
                elif isinstance(external, type):
                    if not is_LikelihoodInterface(external) or \
                            not issubclass(external, Theory):
                        raise LoggedError(self.log, "%s: external class likelihood must "
                                                    "be a subclass of Theory and have "
                                                    "logp, current_logp attributes",
                                          external.__name__)
                    self.add_instance(name, external(info, packages_path=packages_path,
                                                     timing=timing, standalone=False,
                                                     name=name))
                else:
                    # If it has an "external" key, wrap it up. Else, load it up
                    self.add_instance(name, LikelihoodExternalFunction(info, name,
                                                                       timing=timing))
            else:
                assert isinstance(info, Mapping)
                like_class: type = get_component_class(
                    name, kind="likelihood",
                    component_path=info.get("python_path"),
                    class_name=info.get("class"), logger=self.log)
                self.add_instance(name, like_class(info, packages_path=packages_path,
                                                   timing=timing, standalone=False,
                                                   name=name))

            if not is_LikelihoodInterface(self[name]):
                raise LoggedError(self.log, "'Likelihood' %s is not actually a "
                                            "likelihood (no current_logp attribute)",
                                  name)

    def get_helper_theory_collection(self):
        return self.theory

    @property
    def all_types(self):
        if not hasattr(self, "_all_types"):
            self._all_types = set(chain(*[like.type_list for like in self.values()]))
        return self._all_types



from copy import deepcopy
from types import MappingProxyType
from .test_cosmo_planck_2015 import params_lowTEB_highTTTEEE
from .common_cosmo import body_of_test
from cobaya.cosmo_input import planck_lss_precision

best_fit = deepcopy(params_lowTEB_highTTTEEE)
info_camb = MappingProxyType({"camb": {"extra_args": planck_lss_precision["camb"]}})
info_classy = MappingProxyType({"classy": {"extra_args": planck_lss_precision["classy"]}})


def test_cosmo_des_y1_shear_camb(packages_path, skip_not_installed,
                                 info_theory=info_camb):
    like = "des_y1.shear"
    info_likelihood = {like: {}}
    best_fit_shear = deepcopy(best_fit)
    best_fit_shear.update(test_params_shear)
    body_of_test(packages_path, best_fit_shear, info_likelihood, info_theory,
                 {like: ref_chi2["shear"], "tolerance": tolerance},
                 skip_not_installed=skip_not_installed)


def test_cosmo_des_y1_clustering_camb(packages_path, skip_not_installed,
                                      info_theory=info_camb):
    like = "des_y1.clustering"
    info_likelihood = {like: {}}
    best_fit_clustering = deepcopy(best_fit)
    best_fit_clustering.update(test_params_clustering)
    body_of_test(packages_path, best_fit_clustering, info_likelihood, info_theory,
                 {like: ref_chi2["clustering"], "tolerance": tolerance},
                 skip_not_installed=skip_not_installed)


def test_cosmo_des_y1_galaxy_galaxy_camb(packages_path, skip_not_installed,
                                         info_theory=info_camb):
    like = "des_y1.galaxy_galaxy"
    info_likelihood = {like: {}}
    best_fit_galaxy_galaxy = deepcopy(best_fit)
    best_fit_galaxy_galaxy.update(test_params_shear)
    best_fit_galaxy_galaxy.update(test_params_clustering)
    body_of_test(packages_path, best_fit_galaxy_galaxy, info_likelihood, info_theory,
                 {like: ref_chi2["galaxy_galaxy"], "tolerance": tolerance},
                 skip_not_installed=skip_not_installed)


def test_cosmo_des_y1_joint_camb(packages_path, skip_not_installed,
                                 info_theory=info_camb):
    like = "des_y1.joint"
    info_likelihood = {like: {}}
    best_fit_joint = deepcopy(best_fit)
    best_fit_joint.update(test_params_shear)
    best_fit_joint.update(test_params_clustering)
    body_of_test(packages_path, best_fit_joint, info_likelihood, info_theory,
                 {like: ref_chi2["joint"], "tolerance": tolerance},
                 skip_not_installed=skip_not_installed)


def test_cosmo_des_y1_shear_classy(packages_path, skip_not_installed,
                                   info_theory=info_classy):
    test_cosmo_des_y1_shear_camb(packages_path, info_theory=info_theory,
                                 skip_not_installed=skip_not_installed)


def test_cosmo_des_y1_clustering_classy(packages_path, skip_not_installed,
                                        info_theory=info_classy):
    test_cosmo_des_y1_clustering_camb(packages_path, info_theory=info_theory,
                                      skip_not_installed=skip_not_installed)


def test_cosmo_des_y1_galaxy_galaxy_classy(packages_path, skip_not_installed,
                                           info_theory=info_classy):
    test_cosmo_des_y1_galaxy_galaxy_camb(packages_path, info_theory=info_theory,
                                         skip_not_installed=skip_not_installed)


ref_chi2 = {"shear": 242.825, "clustering": 100.78,
            "galaxy_galaxy": 208.005, "joint": 570.428}
tolerance = 0.2

test_params_shear = {
    # wl_photoz_errors
    "DES_DzS1": 0.002,
    "DES_DzS2": -0.015,
    "DES_DzS3": 0.007,
    "DES_DzS4": -0.018,
    # shear_calibration_parameters
    "DES_m1": 0.012,
    "DES_m2": 0.012,
    "DES_m3": 0.012,
    "DES_m4": 0.012,
    # Intrinsic Alignment
    "DES_AIA": 1.0,
    "DES_alphaIA": 1.0,
    "DES_z0IA": 0.62}

test_params_clustering = {
    # lens_photoz_errors
    "DES_DzL1": 0.002,
    "DES_DzL2": 0.001,
    "DES_DzL3": 0.003,
    "DES_DzL4": 0,
    "DES_DzL5": 0,
    # bin_bias
    "DES_b1": 1.45,
    "DES_b2": 1.55,
    "DES_b3": 1.65,
    "DES_b4": 1.8,
    "DES_b5": 2.0}



from cobaya.likelihoods.base_classes import BAO


class sdss_dr16_lrg_bao_dmdh(BAO):
    r"""
    Likelihood of the power spectrum LRG BAO of SDSS DR16 \cite{Alam:2020sor}.
    It does not include the DR12 measurements.
    """
    pass



from copy import deepcopy

from cobaya.component import get_component_class

from .test_cosmo_planck_2015 import params_lowTEB_highTTTEEE, derived_lowTEB_highTTTEEE
from .common_cosmo import body_of_test


# Tests both the bao.generic class, and class renaming for multiple instances
def test_generic_camb(packages_path, skip_not_installed):
    like = "bao.sdss_dr12_consensus_bao"
    like_rename = "my_bao"
    chi2_generic = deepcopy(chi2_sdss_dr12_consensus_bao)
    chi2_generic[like_rename] = chi2_generic.pop(like)
    likelihood_defaults = get_component_class(like).get_defaults()
    likelihood_defaults.pop("path")
    likelihood_defaults["class"] = "bao.generic"
    info_likelihood = {like_rename: likelihood_defaults}
    info_theory = {"camb": None}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_generic, skip_not_installed=skip_not_installed)


def test_sdss_dr16_consensus_bao_lrg_camb(packages_path, skip_not_installed):
    like = "bao.sdss_dr16_baoplus_lrg"
    info_likelihood = {like: {}}
    info_theory = {"camb": None}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_sdss_dr16_baoplus_lrg, skip_not_installed=skip_not_installed)


def test_sdss_dr16_consensus_bao_lrg_classy(packages_path, skip_not_installed):
    like = "bao.sdss_dr16_baoplus_lrg"
    info_likelihood = {like: {}}
    info_theory = {"classy": None}
    chi2 = deepcopy(chi2_sdss_dr16_baoplus_lrg)
    chi2["tolerance"] += chi2.pop("classy_extra_tolerance", 0)
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2, skip_not_installed=skip_not_installed)


def test_sdss_dr16_consensus_bao_elg_camb(packages_path, skip_not_installed):
    like = "bao.sdss_dr16_baoplus_elg"
    info_likelihood = {like: {}}
    info_theory = {"camb": None}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_sdss_dr16_baoplus_elg, skip_not_installed=skip_not_installed)


def test_sdss_dr16_consensus_bao_elg_classy(packages_path, skip_not_installed):
    like = "bao.sdss_dr16_baoplus_elg"
    info_likelihood = {like: {}}
    info_theory = {"classy": None}
    chi2 = deepcopy(chi2_sdss_dr16_baoplus_elg)
    chi2["tolerance"] += chi2.pop("classy_extra_tolerance", 0)
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2, skip_not_installed=skip_not_installed)


def test_sdss_dr16_consensus_bao_qso_camb(packages_path, skip_not_installed):
    like = "bao.sdss_dr16_baoplus_qso"
    info_likelihood = {like: {}}
    info_theory = {"camb": None}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_sdss_dr16_baoplus_qso, skip_not_installed=skip_not_installed)


def test_sdss_dr16_consensus_bao_qso_classy(packages_path, skip_not_installed):
    like = "bao.sdss_dr16_baoplus_qso"
    info_likelihood = {like: {}}
    info_theory = {"classy": None}
    chi2 = deepcopy(chi2_sdss_dr16_baoplus_qso)
    chi2["tolerance"] += chi2.pop("classy_extra_tolerance", 0)
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2, skip_not_installed=skip_not_installed)


def test_sdss_dr16_consensus_bao_lyauto_camb(packages_path, skip_not_installed):
    like = "bao.sdss_dr16_baoplus_lyauto"
    info_likelihood = {like: {}}
    info_theory = {"camb": None}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_sdss_dr16_baoplus_lyauto, skip_not_installed=skip_not_installed)


def test_sdss_dr16_consensus_bao_lyauto_classy(packages_path, skip_not_installed):
    like = "bao.sdss_dr16_baoplus_lyauto"
    info_likelihood = {like: {}}
    info_theory = {"classy": None}
    chi2 = deepcopy(chi2_sdss_dr16_baoplus_lyauto)
    chi2["tolerance"] += chi2.pop("classy_extra_tolerance", 0)
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2, skip_not_installed=skip_not_installed)


def test_sdss_dr16_consensus_bao_lyxqso_camb(packages_path, skip_not_installed):
    like = "bao.sdss_dr16_baoplus_lyxqso"
    info_likelihood = {like: {}}
    info_theory = {"camb": None}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_sdss_dr16_baoplus_lyxqso, skip_not_installed=skip_not_installed)


def test_sdss_dr16_consensus_bao_lyxqso_classy(packages_path, skip_not_installed):
    like = "bao.sdss_dr16_baoplus_lyxqso"
    info_likelihood = {like: {}}
    info_theory = {"classy": None}
    chi2 = deepcopy(chi2_sdss_dr16_baoplus_lyxqso)
    chi2["tolerance"] += chi2.pop("classy_extra_tolerance", 0)
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2, skip_not_installed=skip_not_installed)


def test_sdss_dr12_consensus_bao_camb(packages_path, skip_not_installed):
    like = "bao.sdss_dr12_consensus_bao"
    info_likelihood = {like: {}}
    info_theory = {"camb": None}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_sdss_dr12_consensus_bao, skip_not_installed=skip_not_installed)


def test_sdss_dr12_consensus_bao_classy(packages_path, skip_not_installed):
    like = "bao.sdss_dr12_consensus_bao"
    info_likelihood = {like: {}}
    info_theory = {"classy": None}
    chi2 = deepcopy(chi2_sdss_dr12_consensus_bao)
    chi2["tolerance"] += chi2.pop("classy_extra_tolerance", 0)
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2, skip_not_installed=skip_not_installed)


def test_sdss_dr12_consensus_full_shape_camb(packages_path, skip_not_installed):
    like = "bao.sdss_dr12_consensus_full_shape"
    info_likelihood = {like: {}}
    info_theory = {"camb": None}
    # Check sigma8(z=0): it used to fail bc it was computed internally by the theory code
    # for different redshifts
    derived = {"sigma8": derived_lowTEB_highTTTEEE["sigma8"]}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_sdss_dr12_consensus_full_shape, best_fit_derived=derived,
                 skip_not_installed=skip_not_installed)


def test_sdss_dr12_consensus_full_shape_classy(packages_path, skip_not_installed):
    like = "bao.sdss_dr12_consensus_full_shape"
    info_likelihood = {like: {}}
    info_theory = {"classy": None}
    chi2 = deepcopy(chi2_sdss_dr12_consensus_full_shape)
    chi2["tolerance"] += chi2.pop("classy_extra_tolerance", 0)
    # Check sigma8(z=0): it used to fail bc it was computed internally by the theory code
    # for different redshifts
    derived = {"sigma8": derived_lowTEB_highTTTEEE["sigma8"]}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2, best_fit_derived=derived,
                 skip_not_installed=skip_not_installed)


def test_sdss_dr12_consensus_final_camb(packages_path, skip_not_installed):
    like = "bao.sdss_dr12_consensus_final"
    info_likelihood = {like: {}}
    info_theory = {"camb": None}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_sdss_dr12_consensus_final, skip_not_installed=skip_not_installed)


def test_sdss_dr12_consensus_final_classy(packages_path, skip_not_installed):
    like = "bao.sdss_dr12_consensus_final"
    info_likelihood = {like: {}}
    info_theory = {"classy": None}
    chi2 = deepcopy(chi2_sdss_dr12_consensus_final)
    chi2["tolerance"] += chi2.pop("classy_extra_tolerance", 0)
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2, skip_not_installed=skip_not_installed)


def test_sixdf_2011_bao_camb(packages_path, skip_not_installed):
    like = "bao.sixdf_2011_bao"
    info_likelihood = {like: {}}
    info_theory = {"camb": None}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_sixdf_2011_bao, skip_not_installed=skip_not_installed)


def test_sixdf_2011_bao_classy(packages_path, skip_not_installed):
    like = "bao.sixdf_2011_bao"
    info_likelihood = {like: {}}
    info_theory = {"classy": None}
    chi2 = deepcopy(chi2_sixdf_2011_bao)
    chi2["tolerance"] += chi2.pop("classy_extra_tolerance", 0)
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2, skip_not_installed=skip_not_installed)


def test_sdss_dr7_mgs_camb(packages_path, skip_not_installed):
    like = "bao.sdss_dr7_mgs"
    info_likelihood = {like: {}}
    info_theory = {"camb": None}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_sdss_dr7_mgs, skip_not_installed=skip_not_installed)


def test_sdss_dr7_mgs_classy(packages_path, skip_not_installed):
    like = "bao.sdss_dr7_mgs"
    info_likelihood = {like: {}}
    info_theory = {"classy": None}
    chi2 = deepcopy(chi2_sdss_dr7_mgs)
    chi2["tolerance"] += chi2.pop("classy_extra_tolerance", 0)
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2, skip_not_installed=skip_not_installed)


# BEST FIT AND REFERENCE VALUES ##########################################################

best_fit = deepcopy(params_lowTEB_highTTTEEE)

chi2_sdss_dr16_baoplus_elg = {
    "bao.sdss_dr16_baoplus_elg": 3.24, "tolerance": 0.06}
chi2_sdss_dr16_baoplus_lrg = {
    "bao.sdss_dr16_baoplus_lrg": 5.96, "tolerance": 0.04}
chi2_sdss_dr16_baoplus_qso = {
    "bao.sdss_dr16_baoplus_qso": 8.78, "tolerance": 0.04,
    "classy_extra_tolerance": 0.21}
chi2_sdss_dr16_baoplus_lyauto = {
    "bao.sdss_dr16_baoplus_lyauto": 1.74, "tolerance": 0.04}
chi2_sdss_dr16_baoplus_lyxqso = {
    "bao.sdss_dr16_baoplus_lyxqso": 3.24, "tolerance": 0.04}
chi2_sdss_dr12_consensus_bao = {
    "bao.sdss_dr12_consensus_bao": 5.687, "tolerance": 0.04}
chi2_sdss_dr12_consensus_full_shape = {
    "bao.sdss_dr12_consensus_full_shape": 8.154, "tolerance": 0.02,
    "classy_extra_tolerance": 0.075}
chi2_sdss_dr12_consensus_final = {
    "bao.sdss_dr12_consensus_final": 8.051, "tolerance": 0.03,
    "classy_extra_tolerance": 0.03}
chi2_sixdf_2011_bao = {
    "bao.sixdf_2011_bao": 0.088, "tolerance": 0.02}
chi2_sdss_dr7_mgs = {
    "bao.sdss_dr7_mgs": 0.92689, "tolerance": 0.02}



from cobaya.likelihoods.base_classes import Planck2018Clik


class TTTEEE_unbinned(Planck2018Clik):
    r"""
    Unbinned high-$\ell$ temperature and polarization \textsc{plik} likelihood of Planck's
    2018 data release \cite{Aghanim:2019ame}.
    """
    pass



"""
.. module:: model

:Synopsis: Wrapper for models: parameterization+prior+likelihood+theory
:Author: Jesus Torrado and Antony Lewis

"""
# Global
import os
import dataclasses
from contextlib import contextmanager
from copy import deepcopy
from itertools import chain
from typing import NamedTuple, Sequence, Mapping, Iterable, Optional, \
    Union, List, Any, Dict, Set, Tuple
import numpy as np

# Local
from cobaya.conventions import overhead_time, get_chi2_name, \
    packages_path_input, prior_1d_name
from cobaya.typing import InfoDict, InputDict, LikesDict, TheoriesDict, \
    ParamsDict, PriorsDict, ParamValuesDict, empty_dict, unset_params
from cobaya.input import update_info, load_info_overrides
from cobaya.parameterization import Parameterization
from cobaya.prior import Prior
from cobaya.likelihood import LikelihoodCollection, AbsorbUnusedParamsLikelihood, \
    is_LikelihoodInterface
from cobaya.theory import TheoryCollection, Theory, Provider
from cobaya.log import LoggedError, logger_setup, get_logger, is_debug, HasLogger
from cobaya.yaml import yaml_dump
from cobaya.tools import deepcopy_where_possible, are_different_params_lists, \
    str_to_list, sort_parameter_blocks, recursive_update, sort_cosmetic
from cobaya import mpi


@contextmanager
def timing_on(model: 'Model'):
    was_on = model.timing
    if not was_on:
        model.set_timing_on(True)
    try:
        yield
    finally:
        if not was_on:
            model.set_timing_on(False)


@dataclasses.dataclass(frozen=True)
class LogPosterior:
    """
    Class holding the result of a log-posterior computation, including log-priors,
    log-likelihoods and derived parameters.

    A consistency check will be performed if initialized simultaneously with
    log-posterior, log-priors and log-likelihoods, so, for faster initialisation,
    you may prefer to pass log-priors and log-likelhoods only, and only pass all three
    (so that the test is performed) only when e.g. loading from an old sample.

    If ``finite=True`` (default: False), it will try to represent infinities as the
    largest real numbers allowed by machine precision.
    """

    # A note on typing:
    # Though None is allowed for some arguments, after initialisation everything should
    # be not None. So we can either (a) use Optional, and then get A LOT of typing errors
    # or (b) not use it (use dataclasses.field(default=None) instead) and get fewer errors
    # (only wherever LogPosterior is initialised).
    # Let's opt for (b) and suppress errors there.

    logpost: float = dataclasses.field(default=None)  # type: ignore
    logpriors: Sequence[float] = dataclasses.field(default=None)  # type: ignore
    loglikes: Sequence[float] = dataclasses.field(default=None)  # type: ignore
    derived: Sequence[float] = dataclasses.field(default=None)  # type: ignore
    finite: bool = dataclasses.field(default=False)
    logprior: float = dataclasses.field(init=False, repr=False)
    loglike: float = dataclasses.field(init=False, repr=False)

    def __post_init__(self):
        """Sets the value of logpost if necessary, and checks for consistency."""
        # Consistency: derived = None --> []
        if self.derived is None:
            object.__setattr__(self, 'derived', [])
        object.__setattr__(
            self, 'logprior',
            sum(self.logpriors) if self.logpriors is not None else None)
        object.__setattr__(
            self, 'loglike',
            sum(self.loglikes) if self.loglikes is not None else None)
        if self.finite:
            self.make_finite()
        if self.logpost is None:
            if self.logpriors is None or self.loglikes is None:
                raise ValueError("If `logpost` not passed, both `logpriors` and "
                                 "`loglikes` must be passed.")
            object.__setattr__(self, 'logpost', self._logpost())
        elif self.logpriors is not None and self.loglikes is not None:
            if not self._logpost_is_consistent():
                raise ValueError("The given log-posterior is not equal to the "
                                 "sum of given log-priors and log-likelihoods: "
                                 "%g != sum(%r) + sum(%r)" %
                                 (self.logpost, self.logpriors, self.loglikes))

    def _logpost(self):
        """Computes logpost from prior and likelihood product."""
        return self.logprior + self.loglike

    def _logpost_is_consistent(self):
        """
        Checks that the sum of logpriors and loglikes (if present) add up to logpost, if
        passed.
        """
        if self.finite:
            return np.isclose(np.nan_to_num(self.logpost), np.nan_to_num(self._logpost()))
        else:
            return np.isclose(self.logpost, self._logpost())

    def make_finite(self):
        """
        Ensures that infinities are represented as the largest real numbers allowed by
        machine precision, instead of `+/- numpy.inf`.
        """
        object.__setattr__(self, 'finite', True)
        if self.logpost is not None:
            object.__setattr__(self, 'logpost', np.nan_to_num(self.logpost))
        if self.logpriors is not None:
            object.__setattr__(self, 'logpriors', np.nan_to_num(self.logpriors))
            object.__setattr__(self, 'logprior', np.nan_to_num(self.logprior))
        if self.loglikes is not None:
            object.__setattr__(self, 'loglikes', np.nan_to_num(self.loglikes))
            object.__setattr__(self, 'loglike', np.nan_to_num(self.loglike))

    def as_dict(self, model: "Model") -> Dict[str, Union[float, Dict[str, float]]]:
        """
        Given a :class:`~model.Model`, returns a more informative version of itself,
        containing the names of priors, likelihoods and derived parameters.
        """
        return {
            "logpost": self.logpost,
            "logpriors": dict(zip(model.prior, self.logpriors)),
            "loglikes": dict(zip(model.likelihood, self.loglikes)),
            "derived": dict(zip(model.parameterization.derived_params(), self.derived))}


class Requirement(NamedTuple):
    name: str
    options: Optional[InfoDict]

    def __eq__(self, other):
        return self.name == other.name and _dict_equal(self.options, other.options)

    def __repr__(self):
        return "{%r:%r}" % (self.name, self.options)


def as_requirement_list(requirements):
    if isinstance(requirements, Mapping):
        return [Requirement(name, options) for name, options in requirements.items()]
    elif isinstance(requirements, str):
        return [Requirement(requirements, None)]
    elif isinstance(requirements, Iterable):
        if all(isinstance(term, str) for term in requirements):
            return [Requirement(name, None) for name in requirements]
        result = []
        for item in requirements:
            if isinstance(item, Sequence) and len(item) == 2:
                result.append(Requirement(item[0], item[1]))
            else:
                break
        else:
            return result

    raise ValueError('Requirements must be a dict of names and options, a list of names, '
                     'or an iterable of requirement (name, option) pairs')


def _dict_equal(d1, d2):
    # dict/None equality test accounting for numpy arrays not supporting standard eq
    if type(d1) != type(d2):
        return False
    if isinstance(d1, np.ndarray):
        return np.array_equal(d1, d2)
    if not d1 and not d2:
        return True
    if bool(d1) != bool(d2):
        return False
    if isinstance(d1, str):
        return d1 == d2
    if isinstance(d1, Mapping):
        if set(d1) != set(d2):
            return False
        for k, v in d1.items():
            if not _dict_equal(v, d2[k]):
                return False
        return True
    if hasattr(d1, '__len__'):
        if len(d1) != len(d2):
            return False
        for k1, k2 in zip(d1, d2):
            if not _dict_equal(k1, k2):
                return False
        return True
    return d1 == d2


class Model(HasLogger):
    """
    Class containing all the information necessary to compute the unnormalized posterior.

    Allows for low-level interaction with the theory code, prior and likelihood.

    **NB:** do not initialize this class directly; use :func:`~model.get_model` instead,
    with some info as input.
    """

    provider: Any

    def __init__(self, info_params: ParamsDict, info_likelihood: LikesDict,
                 info_prior: Optional[PriorsDict] = None,
                 info_theory: Optional[TheoriesDict] = None,
                 packages_path=None, timing=None, allow_renames=True, stop_at_error=False,
                 post=False, skip_unused_theories=False,
                 dropped_theory_params: Optional[Iterable[str]] = None):
        """
        Creates an instance of :class:`model.Model`.

        It is recommended to use the simpler function :func:`~model.get_model` instead.
        """
        self.set_logger()
        self._updated_info: InputDict = {
            "params": deepcopy_where_possible(info_params),
            "likelihood": deepcopy_where_possible(info_likelihood)}
        if not self._updated_info["likelihood"]:
            raise LoggedError(self.log, "No likelihood requested!")
        for k, v in (("prior", info_prior), ("theory", info_theory),
                     (packages_path_input, packages_path), ("timing", timing)):
            if v not in (None, {}):
                self._updated_info[k] = deepcopy_where_possible(v)  # type: ignore
        self.parameterization = Parameterization(self._updated_info["params"],
                                                 allow_renames=allow_renames,
                                                 ignore_unused_sampled=post)
        self.prior = Prior(self.parameterization, self._updated_info.get("prior"))
        self.timing = timing
        info_theory = self._updated_info.get("theory")
        self.theory = TheoryCollection(info_theory or {}, packages_path=packages_path,
                                       timing=timing)
        info_likelihood = self._updated_info["likelihood"]
        self.likelihood = LikelihoodCollection(info_likelihood, theory=self.theory,
                                               packages_path=packages_path, timing=timing)
        if stop_at_error:
            for component in self.components:
                component.stop_at_error = stop_at_error
        # Assign input/output parameters
        self._assign_params(info_likelihood, info_theory, dropped_theory_params)
        self._set_dependencies_and_providers(skip_unused_theories=skip_unused_theories)
        # Add to the updated info some values that are only available after initialisation
        self._updated_info = recursive_update(
            self._updated_info, self.get_versions(add_version_field=True))
        # Overhead per likelihood evaluation, approximately ind from # input params
        # Evaluation of non-uniform priors will add some overhead per parameter.
        self.overhead = overhead_time

    def info(self) -> InputDict:
        """
        Returns a copy of the information used to create the model, including defaults
        and some new values that are only available after initialisation.
        """
        return deepcopy_where_possible(self._updated_info)

    def _to_sampled_array(self,
                          params_values: Union[Dict[str, float], Sequence[float]]
                          ) -> np.ndarray:
        """
        Internal method to interact with the prior.
        Needs correct (not renamed) parameter names.
        """
        if hasattr(params_values, "keys"):
            params_values_array = np.array(list(params_values.values()))
        else:
            params_values_array = np.atleast_1d(params_values)
            if params_values_array.shape[0] != self.prior.d():
                raise LoggedError(
                    self.log, "Wrong dimensionality: it's %d, and it should be %d.",
                    len(params_values_array), self.prior.d())
        if len(params_values_array.shape) >= 2:
            raise LoggedError(
                self.log, "Cannot take arrays of points as inputs, just single points.")
        return params_values_array

    def logpriors(self,
                  params_values: Union[Dict[str, float], Sequence[float]],
                  as_dict: bool = False, make_finite: bool = False
                  ) -> Union[np.ndarray, Dict[str, float]]:
        """
        Takes an array or dictionary of sampled parameter values.
        If the argument is an array, parameters must have the same order as in the input.
        When in doubt, you can get the correct order as
        ``list([your_model].parameterization.sampled_params())``.

        Returns the log-values of the priors, in the same order as it is returned by
        ``list([your_model].prior)``. The first one, named ``0``, corresponds to the
        product of the 1-dimensional priors specified in the ``params`` block, and it's
        normalized (in general, the external prior densities aren't).

        If ``as_dict=True`` (default: False), returns a dictionary containing the prior
        names as keys, instead of an array.

        If ``make_finite=True``, it will try to represent infinities as the largest real
        numbers allowed by machine precision.
        """
        params_values = self.parameterization.check_sampled(params_values)
        params_values_array = self._to_sampled_array(params_values)
        logpriors = np.asarray(self.prior.logps(params_values_array))
        if make_finite:
            return np.nan_to_num(logpriors)
        if as_dict:
            return dict(zip(self.prior, logpriors))
        else:
            return logpriors

    def logprior(self,
                 params_values: Union[Dict[str, float], Sequence[float]],
                 make_finite: bool = False) -> float:
        """
        Takes an array or dictionary of sampled parameter values.
        If the argument is an array, parameters must have the same order as in the input.
        When in doubt, you can get the correct order as
        ``list([your_model].parameterization.sampled_params())``.

        Returns the log-value of the prior (in general, unnormalized, unless the only
        priors specified are the 1-dimensional ones in the ``params`` block).

        If ``make_finite=True``, it will try to represent infinities as the largest real
        numbers allowed by machine precision.
        """
        logprior = np.sum(self.logpriors(params_values))
        if make_finite:
            return np.nan_to_num(logprior)
        return logprior

    def _loglikes_input_params(
            self, input_params: Optional[Dict[str, float]] = None,
            return_derived: bool = True, return_output_params: bool = False,
            as_dict: bool = False, make_finite: bool = False, cached: bool = True
    ) -> Union[np.ndarray, Dict[str, float], Tuple[np.ndarray, np.ndarray],
    Tuple[Dict[str, float], Dict[str, float]]]:
        """
        Takes a dict of input parameters, computes the likelihood pipeline, and returns
        the log-likelihoods and derived parameters.

        To return just the list of log-likelihood values, make ``return_derived=False``
        (default: True).

        To return raw output parameters (including chi2's) instead of derived parameters,
        make ``return_output_params=True`` (default=False). This overrides
        ``return_derived``.

        If ``as_dict=True`` (default: False), returns a dictionary containing the
        likelihood names (and derived parameters, if ``return_derived=True``) as keys,
        instead of an array.

        If ``make_finite=True``, it will try to represent infinities as the largest real
        numbers allowed by machine precision.

        If ``cached=False`` (default: True), it ignores previously computed results that
        could be reused.
        """
        # Calculate required results and returns likelihoods
        outpar_dict: ParamValuesDict = {}
        compute_success = True
        self.provider.set_current_input_params(input_params)
        self.log.debug("Got input parameters: %r", input_params)
        loglikes = np.zeros(len(self.likelihood))
        need_derived = self.requires_derived or return_derived or return_output_params
        for (component, like_index), param_dep in zip(self._component_order.items(),
                                                      self._params_of_dependencies):
            depend_list = [input_params[p] for p in param_dep]
            params = {p: input_params[p] for p in component.input_params}
            compute_success = component.check_cache_and_compute(
                params, want_derived=need_derived,
                dependency_params=depend_list, cached=cached)
            if not compute_success:
                loglikes[:] = -np.inf
                self.log.debug("Calculation failed, skipping rest of calculations ")
                break
            if return_derived or return_output_params:
                outpar_dict.update(component.current_derived)
            if like_index is not None:
                try:
                    loglikes[like_index] = float(component.current_logp)  # type: ignore
                except TypeError:
                    raise LoggedError(
                        self.log,
                        "Likelihood %s has not returned a valid log-likelihood, "
                        "but %r instead.", component,
                        component.current_logp)  # type: ignore
        if make_finite:
            loglikes = np.nan_to_num(loglikes)
        return_likes = dict(zip(self.likelihood, loglikes)) if as_dict else loglikes
        if return_derived or return_output_params:
            if not compute_success:
                return_params_names = (
                    self.output_params if return_output_params else self.derived_params)
                if as_dict:
                    return_params = dict(
                        zip(return_params_names, [np.nan] * len(return_params_names)))
                else:
                    return_params = [np.nan] * len(return_params_names)
            else:
                # Add chi2's to derived parameters
                for chi2_name, indices in self._chi2_names:
                    outpar_dict[chi2_name] = -2 * sum(loglikes[i] for i in indices)
                if return_output_params:
                    return_params = (outpar_dict if as_dict
                                     else list(outpar_dict.values()))
                else:  # explicitly derived, instead of output params
                    derived_dict = self.parameterization.to_derived(outpar_dict)
                    self.log.debug("Computed derived parameters: %s", derived_dict)
                    return_params = (derived_dict if as_dict
                                     else list(derived_dict.values()))
            return return_likes, return_params
        return return_likes

    def loglikes(self,
                 params_values: Optional[Union[Dict[str, float], Sequence[float]]] = None,
                 as_dict: bool = False, make_finite: bool = False,
                 return_derived: bool = True, cached: bool = True
                 ) -> Union[np.ndarray, Dict[str, float], Tuple[np.ndarray, np.ndarray],
    Tuple[Dict[str, float], Dict[str, float]]]:
        """
        Takes an array or dictionary of sampled parameter values.
        If the argument is an array, parameters must have the same order as in the input.
        When in doubt, you can get the correct order as
        ``list([your_model].parameterization.sampled_params())``.

        Returns a tuple ``(loglikes, derived_params)``, where ``loglikes`` are the
        log-values of the likelihoods (unnormalized, in general) in the same order as
        it is returned by ``list([your_model].likelihood)``, and ``derived_params``
        are the values of the derived parameters in the order given by
        ``list([your_model].parameterization.derived_params())``.

        To return just the list of log-likelihood values, make ``return_derived=False``
        (default: True).

        If ``as_dict=True`` (default: False), returns a dictionary containing the
        likelihood names (and derived parameters, if ``return_derived=True``) as keys,
        instead of an array.

        If ``make_finite=True``, it will try to represent infinities as the largest real
        numbers allowed by machine precision.

        If ``cached=False`` (default: True), it ignores previously computed results that
        could be reused.
        """
        params_values = self.parameterization.check_sampled(params_values)
        input_params = self.parameterization.to_input(params_values)
        return self._loglikes_input_params(input_params, return_derived=return_derived,
                                           cached=cached, make_finite=make_finite,
                                           as_dict=as_dict)

    def loglike(self,
                params_values: Optional[Union[Dict[str, float], Sequence[float]]] = None,
                make_finite: bool = False, return_derived: bool = True,
                cached: bool = True) -> Union[float, Tuple[float, np.ndarray]]:
        """
        Takes an array or dictionary of sampled parameter values.
        If the argument is an array, parameters must have the same order as in the input.
        When in doubt, you can get the correct order as
        ``list([your_model].parameterization.sampled_params())``.

        Returns a tuple ``(loglike, derived_params)``, where ``loglike`` is the log-value
        of the likelihood (unnormalized, in general), and ``derived_params``
        are the values of the derived parameters in the order given by
        ``list([your_model].parameterization.derived_params())``.
        If the model contains multiple likelihoods, the sum of the loglikes is returned.

        To return just the list of log-likelihood values, make ``return_derived=False``,
        (default: True).

        If ``make_finite=True``, it will try to represent infinities as the largest real
        numbers allowed by machine precision.

        If ``cached=False`` (default: True), it ignores previously computed results that
        could be reused.
        """
        ret_value = self.loglikes(params_values, return_derived=return_derived,
                                  cached=cached, make_finite=make_finite)
        if return_derived:
            return np.sum(ret_value[0]), ret_value[1]
        else:
            return np.sum(ret_value)

    def logposterior(self,
                     params_values: Union[Dict[str, float], Sequence[float]],
                     as_dict: bool = False, make_finite: bool = False,
                     return_derived: bool = True, cached: bool = True,
                     _no_check: bool = False
                     ) -> Union[LogPosterior, dict]:
        """
        Takes an array or dictionary of sampled parameter values.
        If the argument is an array, parameters must have the same order as in the input.
        When in doubt, you can get the correct order as
        ``list([your_model].parameterization.sampled_params())``.

        Returns a :class:`~model.LogPosterior` object (except if ``as_dict=True``, see
        below), with the following fields:

        - ``logpost``: log-value of the posterior.
        - ``logpriors``: log-values of the priors, in the same order as in
          ``list([your_model].prior)``. The first one, corresponds to the
          product of the 1-dimensional priors specified in the ``params``
          block. Except for the first one, the priors are unnormalized.
        - ``loglikes``: log-values of the likelihoods (unnormalized, in general),
          in the same order as in ``list([your_model].likelihood)``.
        - ``derived``: values of the derived parameters in the order given by
          ``list([your_model].parameterization.derived_params())``.

        Only computes the log-likelihood and the derived parameters if the prior is
        non-null (otherwise the fields ``loglikes`` and ``derived`` are empty lists).

        To ignore the derived parameters, make ``return_derived=False`` (default: True).

        If ``as_dict=True`` (default: False), returns a dictionary containing prior names,
        likelihoods names and, if applicable, derived parameters names as keys, instead of
        a :class:`~model.LogPosterior` object.

        If ``make_finite=True``, it will try to represent infinities as the largest real
        numbers allowed by machine precision.

        If ``cached=False`` (default: True), it ignores previously computed results that
        could be reused.
        """
        if _no_check:
            params_values_array = params_values
        else:
            params_values = self.parameterization.check_sampled(params_values)
            params_values_array = self._to_sampled_array(params_values)
            if self.is_debug():
                self.log.debug(
                    "Posterior to be computed for parameters %s",
                    dict(zip(self.parameterization.sampled_params(),
                             params_values_array)))
            if not np.all(np.isfinite(params_values_array)):
                raise LoggedError(
                    self.log, "Got non-finite parameter values: %r",
                    dict(zip(self.parameterization.sampled_params(),
                             params_values_array)))
        # Notice that we don't use the make_finite in the prior call,
        # to correctly check if we have to compute the likelihood
        logpriors_1d = self.prior.logps_internal(params_values_array)
        if logpriors_1d == -np.inf:
            logpriors = [-np.inf] * (1 + len(self.prior.external))
        else:
            input_params = self.parameterization.to_input(params_values_array)
            logpriors = [logpriors_1d]
            if self.prior.external:
                logpriors.extend(self.prior.logps_external(input_params))
        if -np.inf not in logpriors:
            # noinspection PyUnboundLocalVariable
            like = self._loglikes_input_params(input_params,
                                               return_derived=return_derived,
                                               cached=cached, make_finite=make_finite)
            loglikes, derived_sampler = like if return_derived else (like, [])
        else:
            loglikes = []
            derived_sampler = []
        logposterior = LogPosterior(logpriors=logpriors, loglikes=loglikes,
                                    derived=derived_sampler, finite=make_finite)
        if as_dict:
            return logposterior.as_dict(self)
        else:
            return logposterior

    def logpost(self,
                params_values: Union[Dict[str, float], Sequence[float]],
                make_finite: bool = False, cached: bool = True) -> float:
        """
        Takes an array or dictionary of sampled parameter values.
        If the argument is an array, parameters must have the same order as in the input.
        When in doubt, you can get the correct order as
        ``list([your_model].parameterization.sampled_params())``.

        Returns the log-value of the posterior.

        If ``make_finite=True``, it will try to represent infinities as the largest real
        numbers allowed by machine precision.

        If ``cached=False`` (default: True), it ignores previously computed results that
        could be reused.
        """
        return self.logposterior(params_values, make_finite=make_finite,
                                 return_derived=False, cached=cached).logpost

    def get_valid_point(self, max_tries: int, ignore_fixed_ref: bool = False,
                        logposterior_as_dict: bool = False, random_state=None,
                        ) -> Union[Tuple[np.ndarray, LogPosterior],
    Tuple[np.ndarray, dict]]:
        """
        Finds a point with finite posterior, sampled from the reference pdf.

        It will fail if no valid point is found after `max_tries`.

        If `ignored_fixed_ref=True` (default: `False`), fixed reference values will be
        ignored in favor of the full prior, ensuring some randomness for all parameters
        (useful e.g. to prevent caching when measuring speeds).

        Returns (point, LogPosterior(logpost, logpriors, loglikes, derived))

        If ``logposterior_as_dict=True`` (default: False), returns for the log-posterior
        a dictionary containing prior names, likelihoods names and, if applicable, derived
        parameters names as keys, instead of a :class:`~model.LogPosterior` object.
        """
        for loop in range(max(1, max_tries // self.prior.d())):
            initial_point = self.prior.reference(max_tries=max_tries,
                                                 ignore_fixed=ignore_fixed_ref,
                                                 warn_if_no_ref=not loop,
                                                 random_state=random_state)
            results = self.logposterior(initial_point)
            if results.logpost != -np.inf:
                break
        else:
            if self.prior.reference_is_pointlike:
                raise LoggedError(self.log, "The reference point provided has null "
                                            "likelihood. Set 'ref' to a different point "
                                            "or a pdf.")
            raise LoggedError(self.log, "Could not find random point giving finite "
                                        "posterior after %g tries", max_tries)
        if logposterior_as_dict:
            results = results.as_dict(self)
        return initial_point, results

    def dump_timing(self):
        """
        Prints the average computation time of the theory code and likelihoods.

        It's more reliable the more times the likelihood has been evaluated.
        """
        self.likelihood.dump_timing()
        self.theory.dump_timing()

    # Python magic for the "with" statement
    def __enter__(self):
        return self

    def __exit__(self, exception_type=None, exception_value=None, traceback=None):
        self.likelihood.__exit__(exception_type, exception_value, traceback)
        self.theory.__exit__(exception_type, exception_value, traceback)

    def close(self):
        self.__exit__()

    def get_versions(self, add_version_field=False):
        return dict(theory=self.theory.get_versions(add_version_field=add_version_field),
                    likelihood=self.likelihood.get_versions(
                        add_version_field=add_version_field))

    def get_speeds(self, ignore_sub=False):
        return dict(theory=self.theory.get_speeds(ignore_sub=ignore_sub),
                    likelihood=self.likelihood.get_speeds(ignore_sub=ignore_sub))

    def _set_component_order(self, components, dependencies):
        dependence_order: List[Theory] = []
        deps = {p: s.copy() for p, s in dependencies.items()}
        comps = [c for c in components if not isinstance(c, AbsorbUnusedParamsLikelihood)]
        target_length = len(comps)
        _last = 0
        while len(dependence_order) < target_length:
            for component in list(comps):
                if not deps.get(component):
                    dependence_order.append(component)
                    comps.remove(component)
                    for p, dep in deps.items():
                        dep.discard(component)
            if len(dependence_order) == _last:
                raise LoggedError(self.log, "Circular dependency, cannot calculate "
                                            "%r" % comps)
            _last = len(dependence_order)
        likes = list(self.likelihood.values())
        self._component_order = {c: likes.index(c) if c in likes else None
                                 for c in dependence_order}

    def _set_dependencies_and_providers(self, manual_requirements=empty_dict,
                                        skip_unused_theories=False):
        components: List[Theory] = self.components
        direct_param_dependence: Dict[Theory, Set[str]] = {c: set() for c in components}

        def _tidy_requirements(_require, _component=None):
            # take input requirement and return tuples of
            # requirement names and requirement options
            if not _require:
                return []
            _require = as_requirement_list(_require)
            requirements_in_input_params = set(
                req.name for req in _require).intersection(self.input_params)
            if requirements_in_input_params and _component is not None:
                # Save parameters dependence
                direct_param_dependence[_component].update(requirements_in_input_params)
                # requirements that are sampled parameters automatically satisfied
                return [_req for _req in _require
                        if _req.name not in requirements_in_input_params]
            else:
                return _require

        # ## 1. Get the requirements and providers ##
        # requirements of each component
        requirements: Dict[Theory, List[Requirement]] = {}
        # providers of each *available* requirement (requested or not)
        providers: Dict[str, List[Theory]] = {}
        # set of requirement names that may be required derived parameters
        requirements_are_params: Set[str] = set()
        for component in components:
            component.initialize_with_params()
            requirements[component] = \
                _tidy_requirements(component.get_requirements(), component)
            # Component params converted to requirements if not explicitly sampled
            requirements[component] += \
                [Requirement(p, None) for p in (getattr(component, "params", {}) or []) if
                 p not in self.input_params and p not in component.output_params]
            # Gather what this component can provide
            can_provide = (list(component.get_can_provide()) +
                           list(component.get_can_provide_methods()))
            # Parameters that can be provided
            # Corner case: some components can either take some parameters as input OR
            # provide their own calculation of them. Pop those if required as input.
            provide_params = set(p for p in component.get_can_provide_params() if
                                 all(p != req.name for req in requirements[component]))
            provide_params.update(component.output_params)
            requirements_are_params.update(provide_params)
            # Invert to get the provider(s) of each available product/parameter
            for k in chain(can_provide, provide_params):
                providers[k] = providers.get(k, [])
                if component not in providers[k]:
                    providers[k].append(component)
        # Add requirements requested by hand
        manual_theory = Theory(name='_manual')
        if manual_requirements:
            self._manual_requirements = getattr(self, "_manual_requirements", []) \
                                        + _tidy_requirements(manual_requirements)
            requirements[manual_theory] = deepcopy(self._manual_requirements)

        # ## 2. Assign each requirement to a provider ##
        # store requirements assigned to each provider:
        self._must_provide: Dict[Theory, List[Requirement]] = {c: [] for c in components}
        # inverse of the one above, *minus conditional requirements* --
        # this is the one we need to initialise the provider
        requirement_providers = {}
        dependencies: Dict[
            Theory, Set[Theory]] = {}  # set of components of which each component depends
        used_suppliers = set(c for c in components if c.output_params)
        there_are_more_requirements = True
        # temp list of dictionary of requirements for each component
        must_provide = {c: [Requirement(p, None) for p in c.output_params] for c in
                        components}
        while there_are_more_requirements:
            # Check supplier for each requirement, get dependency and must_provide
            for component, requires in requirements.items():
                for requirement in requires:
                    suppliers = providers.get(requirement.name)
                    if not suppliers:
                        # If failed requirement was manually added,
                        # remove from list, or it will still fail in the next call too
                        requirements[manual_theory] = [
                            req for req in requirements.get(manual_theory, [])
                            if req.name != requirement.name]
                        raise LoggedError(self.log,
                                          "Requirement %s of %r is not provided by any "
                                          "component, nor sampled directly",
                                          requirement.name, component)
                    supplier: Optional[Theory]
                    if len(suppliers) == 1:
                        supplier = suppliers[0]
                    else:
                        supplier = None
                        for sup in suppliers:
                            provide = str_to_list(getattr(sup, 'provides', []))
                            if requirement.name in provide:
                                if supplier:
                                    raise LoggedError(
                                        self.log, "more than one component provides %s",
                                        requirement.name)
                                supplier = sup
                        if not supplier:
                            raise LoggedError(
                                self.log, "requirement %s is provided by more than one "
                                          "component: %s. Use 'provides' keyword to "
                                          "specify which provides it",
                                requirement.name, suppliers)
                    if supplier is component:
                        raise LoggedError(
                            self.log, "Component %r cannot provide %s to itself!",
                            component, requirement.name)
                    requirement_providers[requirement.name] = supplier.get_provider()
                    used_suppliers.add(supplier)
                    declared_requirements_for_this_supplier = \
                        self._must_provide[supplier] + must_provide[supplier]
                    if requirement not in declared_requirements_for_this_supplier:
                        must_provide[supplier] += [requirement]
                    dependencies[component] = \
                        dependencies.get(component, set()) | {supplier}
                    # Requirements per component excluding input params
                    # -- saves some overhead in theory.check_cache_and_compute
                    if component is not manual_theory and \
                            requirement.name not in component.input_params and \
                            requirement.name in requirements_are_params:
                        component.input_params_extra.add(requirement.name)
            # tell each component what it must provide, and collect the
            # conditional requirements for those requests
            there_are_more_requirements = False
            for component, requires in requirements.items():
                # empty the list of requirements, since they have already been assigned,
                # and store here new (conditional) ones
                requires[:] = []
                # .get here accounts for the dummy components of manual reqs
                for request in must_provide.get(component) or []:
                    conditional_requirements = \
                        _tidy_requirements(
                            component.must_provide(
                                **{request.name: request.options}), component)
                    self._must_provide[component].append(request)
                    if conditional_requirements:
                        there_are_more_requirements = True
                        requires += conditional_requirements
            # set component compute order and raise error if circular dependence
            self._set_component_order(components, dependencies)
            # TODO: it would be nice that after this loop we, in some way, have
            # component.get_requirements() return the conditional reqs actually used too,
            # or maybe assign conditional used ones to a property?
            # Reset list
            must_provide = {c: [] for c in components}
        # Expunge manual requirements
        requirements.pop(manual_theory, None)
        # Check that unassigned input parameters are at least required by some component
        if self._unassigned_input:
            self._unassigned_input.difference_update(*direct_param_dependence.values())
            if self._unassigned_input:
                unassigned = self._unassigned_input - self.prior.external_dependence
                if unassigned:
                    raise LoggedError(
                        self.log, "Could not find anything to use input parameter(s) %r.",
                        unassigned)
                else:
                    self.mpi_warning("Parameter(s) %s are only used by the prior",
                                     self._unassigned_input)

        unused_theories = set(self.theory.values()) - used_suppliers
        if unused_theories:
            if skip_unused_theories:
                self.mpi_debug('Theories %s do not need to be computed '
                               'and will be skipped', unused_theories)
                for theory in unused_theories:
                    self._component_order.pop(theory, None)
                    components.remove(theory)
            else:
                self.mpi_warning('Theories %s do not appear to be actually used '
                                 'for anything', unused_theories)

        self.mpi_debug("Components will be computed in the order:")
        self.mpi_debug(" - %r" % list(self._component_order))

        def dependencies_of(_component):
            deps = set()
            for c in dependencies.get(_component, []):
                deps.add(c)
                deps.update(dependencies_of(c))
            return deps

        # ## 3. Save dependencies on components and their parameters ##
        self._dependencies = {c: dependencies_of(c) for c in components}
        # this next one is not a dict to save a lookup per iteration
        self._params_of_dependencies: List[Set[str]] \
            = [set() for _ in self._component_order]
        for component, param_dep in zip(self._component_order,
                                        self._params_of_dependencies):
            param_dep.update(direct_param_dependence.get(component) or [])
            for dep in self._dependencies.get(component, []):
                param_dep.update(
                    set(dep.input_params).union(direct_param_dependence.get(dep) or []))
            param_dep -= set(component.input_params)
            if not len(component.input_params) and not param_dep \
                    and component.get_name() != 'one':
                raise LoggedError(
                    self.log, "Component '%r' seems not to depend on any parameters "
                              "(neither directly nor indirectly)", component)
        # Store the input params and components on which each sampled params depends.
        sampled_input_dependence = self.parameterization.sampled_input_dependence()
        sampled_dependence: Dict[str, List[Theory]] = {p: []
                                                       for p in sampled_input_dependence}
        for p, i_s in sampled_input_dependence.items():
            for component in components:
                if p in component.input_params or i_s and \
                        any(p_i in component.input_params for p_i in i_s):
                    sampled_dependence[p].append(component)
                    for comp in components:
                        if comp is not component and \
                                component in self._dependencies.get(comp, []):
                            sampled_dependence[p].append(comp)
        self.sampled_dependence = sampled_dependence
        self.requires_derived: Set[str] = \
            requirements_are_params.intersection(requirement_providers)

        # ## 4. Initialize the provider and pass it to each component ##
        if self.is_debug_and_mpi_root():
            if requirement_providers:
                self.log.debug("Requirements will be calculated by these components:")
                for req, provider in requirement_providers.items():
                    self.log.debug("- %s: %s", req, provider)
            else:
                self.log.debug("No requirements need to be computed")

        self.provider = Provider(self, requirement_providers)
        for component in components:
            component.initialize_with_provider(self.provider)

    def add_requirements(self, requirements):
        """
        Adds quantities to be computed by the pipeline, for testing purposes.
        """
        self._set_dependencies_and_providers(manual_requirements=requirements)

    def requested(self):
        """
        Get all the requested requirements that will be computed.

        :return: dictionary giving list of requirements calculated by
                each component name
        """
        return {"%r" % c: v for c, v in self._must_provide.items() if v}

    def _assign_params(self, info_likelihood, info_theory=None,
                       dropped_theory_params=None):
        """
        Assign input and output parameters to theories and likelihoods, following the
        algorithm explained in :doc:`DEVEL`.
        """
        self.input_params = [p for p in self.parameterization.input_params() if p not in
                             self.parameterization.dropped_param_set()]
        self.output_params = list(self.parameterization.output_params())
        self.derived_params = list(self.parameterization.derived_params())
        input_assign: Dict[str, List[Theory]] = {p: [] for p in self.input_params}
        output_assign: Dict[str, List[Theory]] = {p: [] for p in self.output_params}
        # Go through all components.
        # NB: self.components iterates over likelihoods first, and then theories
        # so unassigned can by default go to theories
        assign_components = [c for c in self.components
                             if not isinstance(c, AbsorbUnusedParamsLikelihood)]
        for assign, option, prefix, derived_param in (
                (input_assign, "input_params", "input_params_prefix", False),
                (output_assign, "output_params", "output_params_prefix", True)):
            agnostic_likes = []
            for component in assign_components:
                if derived_param:
                    required_params = set(str_to_list(getattr(component, "provides", [])))
                else:
                    required_params = set(
                        p for p, v in as_requirement_list(component.get_requirements())
                        # ignore non-params; it's ok if some non-param goes through
                        if v is None)
                # Identify parameters understood by this likelihood/theory
                # 1a. Does it have input/output params list?
                #     (takes into account that for callables, we can ignore elements)
                if getattr(component, option) is not unset_params:
                    for p in getattr(component, option):
                        try:
                            assign[p] += [component]
                        except KeyError:
                            if not derived_param:
                                raise LoggedError(
                                    self.log,
                                    "Parameter '%s' needed as input for '%s', "
                                    "but not provided.", p, component.get_name())
                # 2. Is there a params prefix?
                elif getattr(component, prefix, None) is not None:
                    for p in assign:
                        if p.startswith(getattr(component, prefix)):
                            assign[p] += [component]
                # 3. Does it have a general (mixed) list of params? (set from default)
                # 4. or otherwise required
                elif getattr(component, "params", None) or required_params:
                    if getattr(component, "params", None):
                        for p, options in getattr(component, "params", {}).items():
                            if not isinstance(options, Mapping) and not derived_param or \
                                    isinstance(options, Mapping) and \
                                    options.get('derived', False) is derived_param:
                                if p in assign:
                                    assign[p] += [component]
                    elif component.get_allow_agnostic():
                        agnostic_likes += [component]
                    if required_params:
                        for p in required_params:
                            if p in assign and component not in assign[p]:
                                assign[p] += [component]
                # 5. No parameter knowledge: store as parameter agnostic
                elif component.get_allow_agnostic():
                    agnostic_likes += [component]

            # 6. If parameter not already assigned give to any component that supports it
            #    Input parameters always assigned to any supporting component.
            unassigned = [p for p in assign if not assign[p]]
            for component in assign_components:
                if derived_param:
                    supports_params = component.get_can_provide_params()
                else:
                    supports_params = component.get_can_support_params()
                pars_to_assign = set(supports_params)
                if dropped_theory_params and not is_LikelihoodInterface(component):
                    pars_to_assign.difference_update(dropped_theory_params)
                for p in (unassigned if derived_param else assign):
                    if p in pars_to_assign and component not in assign[p]:
                        assign[p] += [component]
            # Check that there is only one non-knowledgeable element, and assign
            # unused params
            if len(agnostic_likes) > 1 and not all(assign.values()):
                raise LoggedError(
                    self.log, "More than one parameter-agnostic likelihood/theory "
                              "with respect to %s: %r. Cannot decide "
                              "parameter assignments.", option, agnostic_likes)
            elif agnostic_likes:  # if there is only one
                component = agnostic_likes[0]
                for p, assigned in assign.items():
                    if not assigned:
                        assign[p] += [component]

        # If unit likelihood is present, assign all unassigned inputs to it
        for like in self.likelihood.values():
            if isinstance(like, AbsorbUnusedParamsLikelihood):
                for p, assigned in input_assign.items():
                    if not assigned:
                        assigned.append(like)
                break
        # If there are unassigned input params, check later that they are used by
        # *conditional* requirements of a component (and if not raise error)
        self._unassigned_input = set(
            p for p, assigned in input_assign.items() if not assigned).difference(
            chain(*(self.parameterization.input_dependencies.get(p, []) for p, assigned
                    in input_assign.items() if assigned)))

        chi2_names: Dict[str, List[int]] = {}
        # Add aggregated chi2 for likelihood types
        for i, like in enumerate(self.likelihood.values()):
            for tp in like.type_list:
                name = get_chi2_name(tp)
                if name not in chi2_names:
                    chi2_names[name] = []
                chi2_names[name].append(i)
        # Remove aggregated chi2 that may have been picked up by an agnostic component
        for chi2_name in chi2_names:
            output_assign.pop(chi2_name, None)
        # If chi2__like single-likelihood parameters are explicitly requested, include
        # in derived outputs for use in on output derived parameters
        for p in output_assign:
            if p.startswith(get_chi2_name("")):
                like = p[len(get_chi2_name("")):]
                index = list(self.likelihood).index(like)
                if index is None:
                    raise LoggedError(
                        self.log, "Your derived parameters depend on an unknown "
                                  "likelihood: '%s'", like)
                if p in chi2_names:
                    raise LoggedError(
                        self.log, "Your have likelihoods with type labels that are the "
                                  "same as a likelihood", like)
                chi2_names[p] = [index]
                # They may have been already assigned to an agnostic likelihood,
                # so purge first: no "=+"
                output_assign[p] = [self.likelihood[like]]
        self._chi2_names = tuple(chi2_names.items())
        # Check that there are no unassigned parameters (with the exception of aggr chi2)
        if unassigned_output := [p for p, assigned
                                 in output_assign.items() if not assigned]:
            raise LoggedError(
                self.log, "Could not find whom to assign output parameters %r.",
                unassigned_output)
        # Check that output parameters are assigned exactly once
        if multi_assigned_output := {p: assigned for p, assigned in output_assign.items()
                                     if len(assigned) > 1}:
            raise LoggedError(
                self.log,
                "Output params can only be computed by one likelihood/theory, "
                "but some were claimed by more than one: %r.", multi_assigned_output)
        # Finished! Assign and update infos
        for assign, option in ((input_assign, "input_params"),
                               (output_assign, "output_params")):
            for component in self.components:
                setattr(component, option,
                        [p for p, assign in assign.items() if component in assign])
                # Update infos! (helper theory parameters stored in yaml with host)
                inf = (info_likelihood if component in self.likelihood.values() else
                       info_theory)
                inf = inf.get(component.get_name())
                if inf:
                    inf.pop("params", None)
                    inf[option] = component.get_attr_list_with_helpers(option)
        if self.is_debug_and_mpi_root():
            self.log.debug("Parameters were assigned as follows:")
            for component in self.components:
                self.log.debug("- %r:", component)
                self.log.debug("     Input:  %r", component.input_params)
                self.log.debug("     Output: %r", component.output_params)

    @property
    def components(self) -> List[Theory]:
        return list(chain(self.likelihood.values(), self.theory.values()))

    def get_param_blocking_for_sampler(self, split_fast_slow=False, oversample_power=0):
        """
        Separates the sampled parameters in blocks according to the component(s)
        re-evaluation(s) that changing each one of them involves. Then sorts these blocks
        in an optimal way using the speed (i.e. inverse evaluation time in seconds)
        of each component.

        Returns tuples of ``(params_in_block), (oversample_factor)``,
        sorted by descending variation cost-per-parameter.

        Set ``oversample_power`` to some value between 0 and 1 to control the amount of
        oversampling (default: 0 -- no oversampling). If close enough to 1, it chooses
        oversampling factors such that the same amount of time is spent in each block.

        If ``split_fast_slow=True``, it separates blocks in two sets, only the second one
        having an oversampling factor >1. In that case, the oversampling factor should be
        understood as the total one for all the fast blocks.
        """
        # Get a list of components and their speeds
        speeds = {c.get_name(): getattr(c, "speed", -1) for c in self.components}
        # Add overhead to defined ones (positives)
        # and clip undefined ones to the slowest one
        try:
            min_speed = min(speed for speed in speeds.values() if speed > 0)
        except ValueError:
            # No speeds defined <-- empty sequence passed to min
            min_speed = 1
        for comp in speeds:
            speeds[comp] = max(speeds[comp], min_speed)
            # For now, overhead is constant re # params and very small
            speeds[comp] = (speeds[comp] ** -1 + self.overhead) ** -1
        # Compute "footprint"
        # i.e. likelihoods (and theory) that we must recompute when each parameter changes
        footprints = np.zeros((len(self.sampled_dependence), len(speeds)), dtype=int)
        sampled_dependence_names = {k: [c.get_name() for c in v] for
                                    k, v in self.sampled_dependence.items()}
        for i, ls in enumerate(sampled_dependence_names.values()):
            for j, comp in enumerate(speeds):
                footprints[i, j] = comp in ls
        # Group parameters by footprint
        different_footprints = list(set(tuple(row) for row in footprints))
        blocks = [[p for ip, p in enumerate(self.sampled_dependence)
                   if all(footprints[ip] == fp)] for fp in different_footprints]
        # a) Multiple blocks
        if not split_fast_slow:
            i_optimal_ordering, costs, oversample_factors = sort_parameter_blocks(
                blocks, np.array(list(speeds.values()), dtype=float),
                different_footprints, oversample_power=oversample_power)
            blocks_sorted = [blocks[i] for i in i_optimal_ordering]
        # b) 2-block slow-fast separation
        else:
            if len(blocks) == 1:
                raise LoggedError(self.log, "Requested fast/slow separation, "
                                            "but all parameters have the same speed.")
            # First sort them optimally (w/o oversampling)
            i_optimal_ordering, costs, oversample_factors = sort_parameter_blocks(
                blocks, np.array(list(speeds.values()), dtype=float),
                different_footprints, oversample_power=0)
            blocks_sorted = [blocks[i] for i in i_optimal_ordering]
            footprints_sorted = np.array(different_footprints)[list(i_optimal_ordering)]
            # Then, find the split that maxes cost LOG-differences.
            # Since costs are already "accumulated down",
            # we need to subtract those below each one
            costs_per_block = costs - np.concatenate([costs[1:], [0]])
            # Split them so that "adding the next block to the slow ones" has max cost
            log_differences = np.log(costs_per_block[:-1]) - np.log(costs_per_block[1:])
            i_last_slow: int = np.argmax(log_differences)  # type: ignore
            blocks_split = (lambda L: [list(chain(*L[:i_last_slow + 1])),
                                       list(chain(*L[i_last_slow + 1:]))])(blocks_sorted)
            footprints_split = (
                    [np.array(footprints_sorted[:i_last_slow + 1]).sum(axis=0)] +
                    [np.array(footprints_sorted[i_last_slow + 1:]).sum(axis=0)])
            footprints_split = np.clip(np.array(footprints_split), 0, 1)  # type: ignore
            # Recalculate oversampling factor with 2 blocks
            _, _, oversample_factors = sort_parameter_blocks(
                blocks_split, np.array(list(speeds.values()), dtype=float),
                footprints_split, oversample_power=oversample_power)
            # If no oversampling, slow-fast separation makes no sense: warn and set to 2
            if oversample_factors[1] == 1:
                min_factor = 2
                self.mpi_warning(
                    "Oversampling would be trivial due to small speed difference or "
                    "small `oversample_power`. Set to %d.", min_factor)
            # Finally, unfold `oversampling_factors` to have the right number of elements,
            # taking into account that that of the fast blocks should be interpreted as a
            # global one for all of them.
            # NB: the int() below forces the copy of the factors.
            #     Otherwise the yaml_representer prints references to a single object.
            oversample_factors = (
                    [int(oversample_factors[0])] * (1 + i_last_slow) +
                    [int(oversample_factors[1])] * (len(blocks) - (1 + i_last_slow)))
            self.mpi_debug("Doing slow/fast split. The oversampling factors for "
                           "the fast blocks should be interpreted as a global one "
                           "for all of them")
        if self.is_debug_and_mpi_root():
            self.log.debug(
                "Cost, oversampling factor and parameters per block, in optimal order:")
            for c, o, b in zip(costs, oversample_factors, blocks_sorted):
                self.log.debug("* %g : %r : %r", c, o, b)
        return blocks_sorted, oversample_factors

    def check_blocking(self, blocking):
        """
        Checks the correct formatting of the given parameter blocking and oversampling:
        that it consists of tuples `(oversampling_factor, (param1, param2, etc))`, with
        growing oversampling factors, and containing all parameters.

        Returns the input, once checked as ``(blocks), (oversampling_factors)``.

        If ``draggable=True`` (default: ``False``), checks that the oversampling factors
        are compatible with dragging.
        """
        try:
            oversampling_factors, blocks = zip(*list(blocking))
        except:
            raise LoggedError(
                self.log, "Manual blocking not understood. Check documentation.")
        sampled_params = list(self.sampled_dependence)
        check = are_different_params_lists(list(chain(*blocks)), sampled_params)
        duplicate = check.pop("duplicate_A", None)
        missing = check.pop("B_but_not_A", None)
        unknown = check.pop("A_but_not_B", None)
        if duplicate:
            raise LoggedError(
                self.log, "Manual blocking: repeated parameters: %r", duplicate)
        if missing:
            raise LoggedError(
                self.log, "Manual blocking: missing parameters: %r", missing)
        if unknown:
            raise LoggedError(
                self.log, "Manual blocking: unknown parameters: %r", unknown)
        oversampling_factors = np.array(oversampling_factors)
        if (oversampling_factors != np.sort(oversampling_factors)).all():
            self.log.warning(
                "Manual blocking: speed-blocking *apparently* non-optimal: "
                "oversampling factors must go from small (slow) to large (fast).")
        return blocks, oversampling_factors

    def set_cache_size(self, n_states):
        """
        Sets the number of different parameter points to cache for all theories
        and likelihoods.

        :param n_states: number of cached points
        """
        for theory in self.components:
            theory.set_cache_size(n_states)

    def get_auto_covmat(self, params_info=None, random_state=None):
        """
        Tries to get an automatic covariance matrix for the current model and data.

        ``params_info`` should include the set of parameters for which the covmat will be
        searched (default: None, meaning all sampled parameters).
        """
        if params_info is None:
            params_info = self.parameterization.sampled_params_info()
        try:
            for theory in self.theory.values():
                if hasattr(theory, 'get_auto_covmat'):
                    return theory.get_auto_covmat(
                        params_info, self.info()["likelihood"], random_state=random_state)
        except Exception as e:
            self.log.warning("Something went wrong when looking for a covmat: %r", str(e))
            return None

    def set_timing_on(self, on):
        self.timing = on
        for component in self.components:
            component.set_timing_on(on)

    def measure_and_set_speeds(self, n=None, discard=1, max_tries=np.inf,
                               random_state=None):
        """
        Measures the speeds of the different components (theories and likelihoods). To do
        that it evaluates the posterior at `n` points (default: 1 per MPI process, or 3 if
        single process), discarding `discard` points (default: 1) to mitigate possible
        internal caching.

        Stops after encountering `max_tries` points (default: inf) with non-finite
        posterior.
        """
        self.mpi_info("Measuring speeds... (this may take a few seconds)")
        if n is None:
            n = 1 if mpi.more_than_one_process() else 3
        n_done = 0
        with timing_on(self):
            while n_done < int(n) + int(discard):
                point = self.prior.reference(random_state=random_state,
                                             max_tries=max_tries, ignore_fixed=True,
                                             warn_if_no_ref=False)
                if self.loglike(point, cached=False)[0] != -np.inf:
                    n_done += 1
            self.mpi_debug("Computed %d points to measure speeds.", n_done)
            times = [component.timer.get_time_avg() or 0  # type: ignore
                     for component in self.components]
        if mpi.more_than_one_process():
            # average for different points
            times = np.average(mpi.allgather(times), axis=0)
        measured_speeds = [1 / (1e-7 + time) for time in times]
        self.mpi_info('Setting measured speeds (per sec): %r',
                      {component: float("%.3g" % speed) for component, speed in
                       zip(self.components, measured_speeds)})

        for component, speed in zip(self.components, measured_speeds):
            component.set_measured_speed(speed)


class DummyModel:
    """Dummy class for loading chains for post processing."""

    def __init__(self, info_params, info_likelihood, info_prior=None):
        self.parameterization = Parameterization(info_params, ignore_unused_sampled=True)
        self.prior = [prior_1d_name] + list(info_prior or [])
        self.likelihood = list(info_likelihood)


def get_model(info_or_yaml_or_file: Union[InputDict, str, os.PathLike],
              debug: Optional[bool] = None,
              stop_at_error: Optional[bool] = None,
              packages_path: Optional[str] = None,
              override: Optional[InputDict] = None
              ) -> Model:
    """
    Creates a :class:`model.Model`, from Cobaya's input (either as a dictionary, yaml file
    or yaml string). Input fields/options not needed (e.g. ``sampler``, ``output``,
    ``force``, ...) will simply be ignored.

    :param info_or_yaml_or_file: input options dictionary, yaml file, or yaml text
    :param debug: true for verbose debug output, or a specific logging level
    :param packages_path: path where external packages were installed
       (if external dependencies are present).
    :param stop_at_error: stop if an error is raised
    :param override: option dictionary to merge into the input one, overriding settings
       (but with lower precedence than the explicit keyword arguments)
    :return: a :class:`model.Model` instance.

    """
    flags = {packages_path_input: packages_path, "debug": debug,
             "stop_at_error": stop_at_error}
    info = load_info_overrides(info_or_yaml_or_file, override or {}, **flags)
    # MARKED FOR DEPRECATION IN v3.2
    if info.get("debug_file"):
        raise LoggedError("'debug_file' has been deprecated. If you want to "
                          "save the debug output to a file, use 'debug: [filename]'.")
    # END OF DEPRECATION BLOCK
    logger_setup(info.get("debug"))
    # Inform about ignored info keys
    ignored_info = []
    for k in list(info):
        if k not in {"params", "likelihood", "prior", "theory", packages_path_input,
                     "timing", "stop_at_error", "auto_params", "debug"}:
            value = info.pop(k)  # type: ignore
            if value is not None and (not isinstance(value, Mapping) or value):
                ignored_info.append(k)
    # Create the updated input information, including defaults for each component.
    updated_info = update_info(info)
    if ignored_info:
        get_logger(__name__).warning("Ignored blocks/options: %r", ignored_info)
    if is_debug():
        get_logger(__name__).debug(
            "Input info updated with defaults (dumped to YAML):\n%s",
            yaml_dump(sort_cosmetic(updated_info)))
    # Initialize the parameters and posterior
    return Model(updated_info["params"], updated_info["likelihood"],
                 updated_info.get("prior"), updated_info.get("theory"),
                 packages_path=info.get(packages_path_input),
                 timing=updated_info.get("timing"),
                 stop_at_error=info.get("stop_at_error", False))



"""
Automatic tests of the cosmo_external_likelihood example in the documentation,
to make sure it remains up to date.
"""

import os

from .common import process_packages_path
from .conftest import install_test_wrapper

tests_folder = os.path.dirname(os.path.realpath(__file__))
docs_folder = os.path.join(tests_folder, "..", "docs")
docs_src_folder = os.path.join(docs_folder, "src_examples", "cosmo_external_likelihood")
docs_img_folder = os.path.join(docs_folder, "img")


def test_cosmo_docs_likelihood_camb(packages_path, skip_not_installed):
    packages_path = process_packages_path(packages_path)
    # Since we are going to change dirs, make it absolute
    packages_path = os.path.abspath(packages_path)
    # Go to the folder containing the python code
    cwd = os.getcwd()
    try:
        os.chdir(docs_src_folder)
        lines = open(os.path.join(docs_src_folder, "1_fiducial_Cl.py")).readlines()
        for i, line in enumerate(lines):
            if line.startswith("packages_path"):
                lines[i] = "packages_path = '%s'" % packages_path.strip("\'\"")
        globals_example = {}
        install_test_wrapper(skip_not_installed, exec, "\n".join(lines), globals_example)
        exec(open(os.path.join(docs_src_folder, "2_function.py")).read(), globals_example)
        exec(open(os.path.join(docs_src_folder, "3_info_and_plots.py")).read(),
             globals_example)
    except Exception as excpt:
        raise excpt
    finally:
        # Back to the working directory of the tests, just in case
        os.chdir(cwd)



from cobaya.likelihoods.planck_2018_highl_CamSpec2021.planck_2018_CamSpec2021_python import Planck2018CamSpec2021Python


class TTTE(Planck2018CamSpec2021Python):
    r"""
    High-$\ell$ TT + TE \textsc{CamSpec} 12.6HMcl likelihood using Planck 2018 data
    \cite{efstathiou2019}.
    """
    pass



from cobaya.likelihoods.base_classes import BAO


class sdss_dr16_baoplus_lyauto(BAO):
    r"""
    Likelihood of the power spectrum Lyman alpha BAO of SDSS DR16 \cite{Alam:2020sor}.
    """
    pass



# Global
from copy import deepcopy
from typing import MutableMapping, Mapping

# Local
from cobaya.input import get_default_info, merge_info
from cobaya.parameterization import reduce_info_param
from cobaya.typing import InputDict
from . import input_database


def translate(p, info=None, dictionary=None):
    dictionary = dictionary or {}
    # Ignore if dropped
    if not (info if isinstance(info, Mapping) else {}).get("drop", False):
        p = dictionary.get(p, p)
    # Try to modify lambda parameters too!
    if isinstance(info, str):
        if info.startswith("lambda"):
            arguments = "".join(info.split(":")[0].split()[1:]).split(",")
            arguments_t = [translate(pi, dictionary=dictionary)[0] for pi in arguments]
            for pi, pit in zip(arguments, arguments_t):
                info = info.replace(pi, pit)
    if ((isinstance(info, MutableMapping) and "derived" in info and
         isinstance(info["derived"], str))):
        info["derived"] = translate(p, info["derived"],
                                    dictionary=dictionary)[1]
    elif (isinstance(info, MutableMapping) and "value" in info and
          isinstance(info["value"], str)):
        info["value"] = translate(p, info["value"], dictionary=dictionary)[1]
    return p, info


def create_input(**kwargs) -> InputDict:
    get_comments = kwargs.pop("get_comments", False)
    preset = kwargs.pop("preset", None)
    if preset:
        fields = deepcopy(input_database.preset[preset])
        fields.update(kwargs)
        return create_input(get_comments=get_comments, **fields)
    # Need to copy to select theory.
    # Likelihoods always AT THE END!
    # (to check that sampled parameters are not redefined as derived)
    infos: dict = {}
    kwargs_params = ["primordial", "geometry", "hubble", "matter", "neutrinos",
                     "dark_energy", "bbn", "reionization"]
    kwargs_likes = ["like_cmb", "like_bao", "like_des", "like_sn", "like_H0"]
    for k in kwargs_params + kwargs_likes:
        if k not in kwargs:
            infos[k] = {}
        try:
            infos[k] = deepcopy(
                getattr(input_database, k)[kwargs.get(k, input_database.none)])
        except KeyError:
            raise ValueError("Unknown value '%s' for '%s'" %
                             (kwargs.get(k, input_database.none), k))
    theory_requested = kwargs.get("theory")
    for i, (field, info) in enumerate(infos.items()):
        if not info:
            continue
        error_msg = info.pop(input_database.error_msg, None)
        try:
            info["theory"] = {theory_requested: info["theory"][theory_requested]}
        except KeyError:
            return ("There is no preset for\n'%s'" % (
                info.get("desc", field)) +
                    "with theory code '%s'." % theory_requested +
                    ("\n--> " + error_msg if error_msg else
                     "\nThis does not mean that you cannot use this model with that "
                     "theory code; just that we have not implemented this yet."))
        # Add non-translatable parameters (in info["theory"][classy|camb][params])
        if "params" not in info:
            info["params"] = {}
        info["params"].update(
            (info["theory"][theory_requested] or {}).pop("params", {}))
        # Remove the *derived* parameters mentioned by the likelihoods that
        # are already *sampled* by some part of the model
        if field.startswith("like_") and "params" in info:
            remove_derived = []
            for p in info["params"]:
                if any((p in info_part["params"])
                       for info_part in list(infos.values())[:i]):
                    remove_derived += [p]
            for p in remove_derived:
                info["params"].pop(p)
    # Prepare sampler info
    info_sampler = deepcopy(input_database.sampler.get(kwargs.get("sampler") or "", {}))
    if info_sampler:
        sampler_name = list(info_sampler["sampler"])[0]
        info_sampler["sampler"][sampler_name] = info_sampler["sampler"][
                                                    sampler_name] or {}
        # Add recommended options for samplers
        for info in infos.values():
            this_info_sampler = info.pop("sampler", {})
            if sampler_name in this_info_sampler:
                info_sampler["sampler"][sampler_name].update(
                    this_info_sampler[sampler_name])
    # Reorder, to have the *parameters* shown in the correct order, and *merge*
    all_infos = [info_sampler] + [infos[k]
                                  for k in kwargs_likes[::-1] + kwargs_params[::-1]]
    comments = [info.pop("comment", None) for info in all_infos]
    comments = [c for c in comments if c]
    [info.pop("desc", None) for info in all_infos]
    merged = merge_info(*all_infos)
    # Simplify parameter infos
    for p, info in merged["params"].items():
        merged["params"][p] = reduce_info_param(info)
    # Translate from Planck param names
    planck_to_theo = \
        get_default_info(theory_requested, "theory")["renames"]
    if kwargs.get("planck_names", False):
        merged["theory"][theory_requested] = merged["theory"][
                                                 theory_requested] or {}
        merged["theory"][theory_requested]["use_renames"] = True
    else:
        merged_params_translated = dict(
            translate(p, info, planck_to_theo)
            for p, info in merged["params"].items())
        merged["params"] = merged_params_translated
    if get_comments and comments:
        merged["comment"] = comments
    return merged



r"""
.. module:: samplers.Minimize

:Synopsis: Posterior/likelihood *maximization* (i.e. -log(post) and chi^2 minimization).
:Author: Jesus Torrado

This is a **maximizer** for posteriors or likelihoods, based on
`scipy.optimize.Minimize <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_
and `Py-BOBYQA <https://numericalalgorithmsgroup.github.io/pybobyqa/build/html/index.html>`_.

The default is BOBYQA, which tends to work better on Cosmological problems with default
settings.

.. |br| raw:: html

   <br />

.. note::
   **If you use BOBYQA, please cite it as:**
   |br|
   `C. Cartis, J. Fiala, B. Marteau, L. Roberts,
   "Improving the Flexibility and Robustness of Model-Based Derivative-Free Optimization Solvers"
   (arXiv:1804.00154) <https://arxiv.org/abs/1804.00154>`_
   |br|
   `C. Cartis, L. Roberts, O. Sheridan-Methven,
   "Escaping local minima with derivative-free methods: a numerical investigation"
   (arXiv:1812.11343) <https://arxiv.org/abs/1812.11343>`_
   |br|
   `M.J.D. Powell,
   "The BOBYQA Algorithm for Bound Constrained Optimization without Derivatives",
   (Technical Report 2009/NA06, DAMTP, University of Cambridge)
   <http://www.damtp.cam.ac.uk/user/na/NA_papers/NA2009_06.pdf>`_

   **If you use scipy**, you can find `the appropriate references here
   <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_.

It works more effectively when run on top of a Monte Carlo sample: it will use the maximum
a posteriori as a starting point (or the best fit, depending on whether the prior is
ignored, :ref:`see below <minimize_like>`), and the recovered covariance matrix of the
posterior to rescale the variables.

To take advantage of a previous run with a Monte Carlo sampler, either:

- change the ``sampler`` to ``minimize`` in the input file,

- or, if running from the shell, repeat the ``cobaya-run`` command used for the original
  run, adding the ``--minimize`` flag.

When called from a Python script, Cobaya's ``run`` function returns the updated info
and the products described below in the method
:func:`samplers.minimize.Minimize.products` (see below).

If text output is requested, it produces two different files:

- ``[output prefix].minimum.txt``, in
  :ref:`the same format as Cobaya samples <output_format>`,
  but containing a single line.

- ``[output prefix].minimum``, the equivalent **GetDist-formatted** file.

.. warning::

   For historical reasons, in the first two lines of the GetDist-formatted output file
   ``-log(Like)`` indicates the negative log-**posterior**, and similarly ``chi-sq`` is
   :math:`-2` times the log-**posterior**. The actual log-likelihood can be obtained as
   :math:`-2` times the sum of the individual :math:`\chi^2` (``chi2__``, with double
   underscore) in the table that follows these first lines.

It is recommended to run a couple of parallel MPI processes:
it will finally pick the best among the results.

.. warning::

   Since Cobaya is often used on likelihoods featuring numerical noise (e.g. Cosmology),
   we have reduced the default accuracy criterion for the minimizers, so that they
   converge in a limited amount of time. If your posterior is fast to evaluate, you may
   want to refine the convergence parameters (see ``override`` options in the ``yaml``
   below).


.. _minimize_like:

Maximizing the likelihood instead of the posterior
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

To maximize the likelihood, add ``ignore_prior: True`` in the ``minimize`` input block.

When producing text output, the generated files are named ``.bestfit[.txt]`` instead of
``minimum``, and contain the best-fit (maximum of the likelihood) instead of the MAP
(maximum of the posterior).

"""

# Global
import os
import re
from itertools import chain
from typing import Optional, Union
import numpy as np
from scipy import optimize
import pybobyqa
from pybobyqa import controller

# Local
from cobaya.sampler import Minimizer
from cobaya.conventions import undo_chi2_name
from cobaya.collection import OnePoint, SampleCollection
from cobaya.log import LoggedError
from cobaya.tools import read_dnumber, recursive_update
from cobaya.sampler import CovmatSampler
from cobaya import mpi

# Handling scipy vs BOBYQA
evals_attr = {"scipy": "fun", "bobyqa": "f"}
valid_methods = tuple(evals_attr)

# Conventions conventions
getdist_ext_ignore_prior = {True: ".bestfit", False: ".minimum"}
get_collection_extension = (
    lambda ignore_prior: getdist_ext_ignore_prior[ignore_prior] + ".txt")

_bobyqa_errors = {
    controller.EXIT_MAXFUN_WARNING:
        "Maximum allowed objective evaluations reached. "
        "This is the most likely return value when using multiple restarts.",
    controller.EXIT_SLOW_WARNING:
        "Maximum number of slow iterations reached.",
    controller.EXIT_FALSE_SUCCESS_WARNING:
        "Py-BOBYQA reached the maximum number of restarts which decreased the"
        " objective, but to a worse value than was found in a previous run.",
    controller.EXIT_INPUT_ERROR:
        "Error in the inputs.",
    controller.EXIT_TR_INCREASE_ERROR:
        "Error occurred when solving the trust region subproblem.",
    controller.EXIT_LINALG_ERROR:
        "Linear algebra error, e.g. the interpolation points produced a "
        "singular linear system."}


class Minimize(Minimizer, CovmatSampler):
    file_base_name = 'minimize'

    ignore_prior: bool
    confidence_for_unbounded: float
    method: str
    best_of: int
    override_bobyqa: Optional[dict]
    override_scipy: Optional[dict]
    max_evals: Union[str, int]

    def initialize(self):
        """
        Initializes the minimizer: sets the boundaries of the problem, selects starting
        points and sets up the affine transformation.
        """
        if self.method not in evals_attr:
            raise LoggedError(self.log, "Method '%s' not recognized. Try one of %r.",
                              self.method, list(evals_attr))
        self.mpi_info("Initializing")
        self.max_iter = int(read_dnumber(self.max_evals, self.model.prior.d()))
        # Configure target
        method = self.model.loglike if self.ignore_prior else self.model.logpost
        kwargs = {"make_finite": True}
        if self.ignore_prior:
            kwargs["return_derived"] = False
        self.logp = lambda x: method(x, **kwargs)
        # Try to load info from previous samples.
        # If none, sample from reference (make sure that it has finite like/post)
        self.initial_points = []
        assert self.best_of > 0
        num_starts = int(np.ceil(self.best_of / mpi.size()))
        if self.output:
            files = self.output.find_collections()
        else:
            files = None
        for start in range(num_starts):
            initial_point = None
            if files:
                collection_in: Optional[SampleCollection]
                if mpi.more_than_one_process() or num_starts > 1:
                    index = 1 + mpi.rank() * num_starts + start
                    if index <= len(files):
                        collection_in = SampleCollection(
                            self.model, self.output, name=str(index), resuming=True)
                    else:
                        collection_in = None
                else:
                    collection_in = self.output.load_collections(self.model,
                                                                 concatenate=True)
                if collection_in:
                    initial_point = (collection_in.bestfit() if self.ignore_prior
                                     else collection_in.MAP())
                    initial_point = initial_point[
                        list(self.model.parameterization.sampled_params())].values
                    self.log.info(
                        "Run %d/%d will start from %s of previous sample:",
                        start + 1, num_starts, "best fit" if self.ignore_prior else "MAP"
                    )
                    # Compute covmat if input but no .covmat file (e.g. with PolyChord)
                    # Prefer old over `covmat` definition in yaml (same as MCMC)
                    self.covmat = collection_in.cov(derived=False)
                    self.covmat_params = list(
                        self.model.parameterization.sampled_params())
            if initial_point is None:
                for _ in range(self.max_iter // 10 + 5):
                    initial_point = self.model.prior.reference(random_state=self._rng)
                    if np.isfinite(self.logp(initial_point)):
                        break
                else:
                    raise LoggedError(self.log, "Could not find random starting point "
                                                "giving finite posterior")
                self.log.info(
                    "Run %d/%d will start from random initial point:",
                    start + 1, num_starts
                )
            self.log.info(
                dict(zip(self.model.parameterization.sampled_params(), initial_point)))
            self.initial_points.append(initial_point)
        self._bounds = self.model.prior.bounds(
            confidence_for_unbounded=self.confidence_for_unbounded)
        # TODO: if ignore_prior, one should use *like* covariance (this is *post*)
        covmat = self._load_covmat(prefer_load_old=self.output)[0]
        # scale by conditional parameter widths (since not using correlation structure)
        scales = np.minimum(1 / np.sqrt(np.diag(np.linalg.inv(covmat))),
                            (self._bounds[:, 1] - self._bounds[:, 0]) / 3)
        # Cov and affine transformation
        # Transform to space where initial point is at centre, and cov is normalised
        # Cannot do rotation, as supported minimization routines assume bounds aligned
        # with the parameter axes.
        self._affine_transform_matrix = np.diag(1 / scales)
        self._inv_affine_transform_matrix = np.diag(scales)
        self._affine_transform_baseline = None
        self._scales = scales
        self.kwargs = None
        self.result = None
        self.minimum = None

    def affine_transform(self, x):
        """Transforms a point into the search space."""
        return (x - self._affine_transform_baseline) / self._scales

    def inv_affine_transform(self, x):
        """Transforms a point from the search space back into the parameter space."""
        # Fix up rounding errors on bounds to avoid -np.inf likelihoods
        return np.clip(x * self._scales + self._affine_transform_baseline,
                       self._bounds[:, 0], self._bounds[:, 1])

    def run(self):
        """
        Runs `scipy.Minimize`
        """
        results = []
        successes = []

        def minuslogp_transf(x):
            return -self.logp(self.inv_affine_transform(x))

        for i, initial_point in enumerate(self.initial_points):
            self.log.info("Starting run %d/%d", i + 1, len(self.initial_points))
            self.log.debug("Starting point: %r", initial_point)
            self._affine_transform_baseline = initial_point
            initial_point = self.affine_transform(initial_point)
            np.testing.assert_allclose(initial_point, np.zeros(initial_point.shape))
            bounds = np.array(
                [self.affine_transform(self._bounds[:, i]) for i in range(2)]).T
            try:
                # Configure method
                if self.method.lower() == "bobyqa":
                    self.kwargs = {
                        "objfun": minuslogp_transf,
                        "x0": initial_point,
                        "bounds": np.array(list(zip(*bounds))),
                        "maxfun": self.max_iter,
                        "rhobeg": 1.,
                        "do_logging": self.is_debug()}
                    self.kwargs = recursive_update(self.kwargs,
                                                   self.override_bobyqa or {})
                    self.log.debug("Arguments for pybobyqa.solve:\n%r",
                                   {k: v for k, v in self.kwargs.items() if
                                    k != "objfun"})
                    result = pybobyqa.solve(**self.kwargs)
                    success = result.flag == result.EXIT_SUCCESS
                    if not success:
                        self.log.error(
                            "Finished unsuccessfully. Reason: %s",
                            _bobyqa_errors[result.flag]
                        )
                else:
                    self.kwargs = {
                        "fun": minuslogp_transf,
                        "x0": initial_point,
                        "bounds": bounds,
                        "options": {
                            "maxiter": self.max_iter,
                            "disp": self.is_debug()}}
                    self.kwargs = recursive_update(self.kwargs, self.override_scipy or {})
                    self.log.debug("Arguments for scipy.optimize.Minimize:\n%r",
                                   {k: v for k, v in self.kwargs.items() if k != "fun"})
                    result = optimize.minimize(**self.kwargs)
                    success = result.success
                    if not success:
                        self.log.error("Finished unsuccessfully.")
                if success:
                    self.log.info(
                        "Run %d/%d converged.", i + 1, len(self.initial_points)
                    )
            except Exception as excpt:
                self.log.error("Minimizer '%s' raised an unexpected error:", self.method)
                raise excpt
            results += [result]
            successes += [success]
        self.process_results(*mpi.zip_gather(
            [results, successes, self.initial_points,
             [self._inv_affine_transform_matrix] * len(self.initial_points)]))

    @mpi.set_from_root(("_inv_affine_transform_matrix", "_affine_transform_baseline",
                        "result", "minimum"))
    def process_results(self, results, successes, affine_transform_baselines,
                        transform_matrices):
        """
        Determines success (or not), chooses best (if MPI or multiple starts)
        and produces output (if requested).
        """
        evals_attr_ = evals_attr[self.method.lower()]
        results = list(chain(*results))
        successes = list(chain(*successes))
        affine_transform_baselines = list(chain(*affine_transform_baselines))
        transform_matrices = list(chain(*transform_matrices))
        if len(results) > 1:
            mins = [(getattr(r, evals_attr_) if s else np.inf)
                    for r, s in zip(results, successes)]
            i_min: int = np.argmin(mins)  # type: ignore
        else:
            i_min = 0
        self.result = results[i_min]
        self._affine_transform_baseline = affine_transform_baselines[i_min]
        self._inv_affine_transform_matrix = transform_matrices[i_min]
        if not any(successes):
            raise LoggedError(
                self.log, "Minimization failed! Here is the raw result object:\n%s",
                str(self.result))
        elif not all(successes):
            self.log.warning('Some minimizations failed!')
        elif len(results) > 1:
            self.log.info('Finished successfully!')
            # noinspection PyUnboundLocalVariable
            if max(mins) - min(mins) > 1:
                self.log.warning('Big spread in minima: %r', mins)
            elif max(mins) - min(mins) > 0.2:
                self.log.warning('Modest spread in minima: %r', mins)
        logp_min = -np.array(getattr(self.result, evals_attr_))
        x_min = self.inv_affine_transform(self.result.x)
        self.log.info("-log(%s) minimized to %g",
                      "likelihood" if self.ignore_prior else "posterior", -logp_min)
        recomputed_post_min = self.model.logposterior(x_min, cached=False)
        recomputed_logp_min = (recomputed_post_min.loglike if self.ignore_prior
                               else recomputed_post_min.logpost)
        if not np.allclose(logp_min, recomputed_logp_min, atol=1e-2):
            raise LoggedError(
                self.log, "Cannot reproduce log minimum to within 0.01. Maybe your "
                          "likelihood is stochastic or large numerical error? "
                          "Recomputed min: %g (was %g) at %r",
                recomputed_logp_min, logp_min, x_min)
        self.minimum = OnePoint(self.model, self.output, name="",
                                extension=get_collection_extension(self.ignore_prior))
        self.minimum.add(x_min, derived=recomputed_post_min.derived,
                         logpost=recomputed_post_min.logpost,
                         logpriors=recomputed_post_min.logpriors,
                         loglikes=recomputed_post_min.loglikes)
        self.log.info(
            "Parameter values at minimum:\n%s", self.minimum.data.to_string())
        self.minimum.out_update()
        self.dump_getdist()

    def products(self):
        r"""
        Returns a dictionary containing:

        - ``minimum``: :class:`OnePoint` that maximizes the posterior or likelihood
          (depending on ``ignore_prior``).

        - ``result_object``: instance of results class of
          `scipy <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.OptimizeResult.html>`_
          or `pyBOBYQA
          <https://numericalalgorithmsgroup.github.io/pybobyqa/build/html/userguide.html>`_.

        - ``M``: inverse of the affine transform matrix (see below).
          ``None`` if no transformation applied.

        - ``X0``: offset of the affine transform matrix (see below)
          ``None`` if no transformation applied.

        If non-trivial ``M`` and ``X0`` are returned, this means that the minimizer has
        been working on an affine-transformed parameter space :math:`x^\prime`, from which
        the real space points can be obtained as :math:`x = M x^\prime + X_0`. This inverse
        transformation needs to be applied to the coordinates appearing inside the
        ``result_object``.
        """
        return {"minimum": self.minimum, "result_object": self.result,
                "M": self._inv_affine_transform_matrix,
                "X0": self._affine_transform_baseline}

    def getdist_point_text(self, params, weight=None, minuslogpost=None):
        """Creates the multi-line string containing the minumum in GetDist format."""
        lines = []
        if weight is not None:
            lines.append('  weight    = %s' % weight)
        if minuslogpost is not None:
            lines.append(' -log(Like) = %s' % minuslogpost)
            lines.append('  chi-sq    = %s' % (2 * minuslogpost))
        lines.append('')
        labels = self.model.parameterization.labels()
        label_list = list(labels)
        if hasattr(params, 'chi2_names'):
            label_list += params.chi2_names
        width = max(len(lab) for lab in label_list) + 2

        def add_section(pars):
            for p, val in pars:
                lab = labels.get(p, p)
                num = label_list.index(p) + 1
                if isinstance(val, (float, np.floating)) and len(str(val)) > 10:
                    lines.append("%5d  %-17.9e %-*s %s" % (num, val, width, p, lab))
                else:
                    lines.append("%5d  %-17s %-*s %s" % (num, val, width, p, lab))
        add_section(
            [(p, params[p]) for p in self.model.parameterization.sampled_params()])
        lines.append('')
        add_section([[p, value] for p, value in
                     self.model.parameterization.constant_params().items()])
        lines.append('')
        add_section(
            [[p, params[p]] for p in self.model.parameterization.derived_params()])
        if hasattr(params, 'chi2_names'):
            labels.update(
                {p: r'\chi^2_{\rm %s}' % (
                    undo_chi2_name(p).replace("_", r"\ "))
                 for p in params.chi2_names}
            )
            add_section([[chi2, params[chi2]] for chi2 in params.chi2_names])
        return "\n".join(lines)

    def dump_getdist(self):
        """Writes the GetDist format point."""
        if not self.output:
            return
        getdist_bf = self.getdist_point_text(self.minimum,
                                             minuslogpost=self.minimum['minuslogpost'])
        out_filename = os.path.join(
            self.output.folder,
            self.output.prefix + getdist_ext_ignore_prior[self.ignore_prior])
        with open(out_filename, 'w', encoding="utf-8") as f:
            f.write(getdist_bf)

    @classmethod
    def output_files_regexps(cls, output, info=None, minimal=False):
        """
        Returns a list of tuples `(regexp, root)` of output files potentially produced.
        If `root` in the tuple is `None`, `output.folder` is used.

        If `minimal=True`, returns regexp's for the files that should really not be there
        when we are not resuming.
        """
        ignore_prior = bool(info.get("ignore_prior", False))
        ext_collection = get_collection_extension(ignore_prior)
        ext_getdist = getdist_ext_ignore_prior[ignore_prior]
        regexps = [
            re.compile(output.prefix_regexp_str + re.escape(ext.lstrip(".")) + "$")
            for ext in [ext_collection, ext_getdist]]
        return [(r, None) for r in regexps]

    @classmethod
    def check_force_resume(cls, output, info=None):
        """
        Performs the necessary checks on existing files if resuming or forcing
        (including deleting some output files when forcing).
        """
        if output.is_resuming():
            if mpi.is_main_process():
                raise LoggedError(
                    output.log, "Minimizer does not support resuming. "
                                "If you want to start over, force "
                                "('-f', '--force', 'force: True')")
        super().check_force_resume(output, info=info)

    @classmethod
    def _get_desc(cls, info=None):
        if info is None:
            method = None
        else:
            method = info.get("method", cls.get_defaults()["method"])
        desc_bobyqa = (r"Py-BOBYQA implementation "
                       r"\cite{2018arXiv180400154C,2018arXiv181211343C} of the BOBYQA "
                       r"minimization algorithm \cite{BOBYQA}")
        desc_scipy = (r"Scipy minimizer \cite{2020SciPy-NMeth} (check citation for the "
                      r"actual algorithm used at \url{https://docs.scipy.org/doc/scipy/re"
                      r"ference/generated/scipy.optimize.Minimize.html}")
        if method and method.lower() == "bobyqa":
            return desc_bobyqa
        elif method and method.lower() == "scipy":
            return desc_scipy
        else:  # unknown method or no info passed (None)
            return ("Minimizer -- method unknown, possibly one of:"
                    "\na) " + desc_bobyqa + "\nb) " + desc_scipy)



"""
.. module:: likelihoods.gaussian_mixture

:Synopsis: Gaussian mixture likelihood
:Author: Jesus Torrado

"""
# Global
from typing import Any
import numpy as np
from scipy.stats import multivariate_normal, uniform, random_correlation
from scipy.special import logsumexp

# Local
from cobaya.likelihood import Likelihood
from cobaya.log import LoggedError
from cobaya.mpi import share_mpi, is_main_process
from cobaya.typing import InputDict, Union, Sequence

derived_suffix = "_derived"


class GaussianMixture(Likelihood):
    """
    Gaussian likelihood.
    """
    file_base_name = 'gaussian_mixture'

    # yaml variables
    means: Union[Sequence, np.ndarray]
    covs: Union[Sequence, np.ndarray]
    weights: Union[np.ndarray, float]
    derived: bool
    input_params_prefix: str
    output_params_prefix: str

    def d(self):
        """
        Dimension of the input vector.
        """
        return len(self.input_params)

    def initialize_with_params(self):
        """
        Initializes the gaussian distributions.
        """
        self.log.debug("Initializing")
        # Load mean and cov, and check consistency of n_modes and dimensionality
        if self.means is not None and self.covs is not None:
            # Wrap them in the right arrays (n_mode, param) and check consistency
            self.means = np.atleast_1d(self.means)
            while len(np.array(self.means).shape) < 2:
                self.means = np.array([self.means])
            mean_n_modes, mean_dim = self.means.shape
            self.covs = np.atleast_1d(self.covs)
            while len(np.array(self.covs).shape) < 3:
                self.covs = np.array([self.covs])
            cov_n_modes, cov_dim, cov_dim_2 = self.covs.shape
            if cov_dim != cov_dim_2:
                raise LoggedError(
                    self.log, "The covariance matrix(/ces) do not appear to be square!\n"
                              "Got %r", self.covs)
            if mean_dim != cov_dim:
                raise LoggedError(
                    self.log,
                    "The dimensionalities guessed from mean(s) and cov(s) do not match!")
            if mean_n_modes != cov_n_modes:
                raise LoggedError(
                    self.log,
                    "The numbers of modes guessed from mean(s) and cov(s) do not match!")
            if mean_dim != self.d():
                raise LoggedError(
                    self.log,
                    "The dimensionality is %d (guessed from given means and covmats) "
                    "but was passed %d parameters instead. " +
                    ("Maybe you forgot to specify the prefix by which to identify them?"
                     if self.input_params_prefix else ""), mean_dim,
                    len(self.input_params))
            self.n_modes = mean_n_modes
            if self.derived and len(self.output_params) != self.d() * self.n_modes:
                raise LoggedError(
                    self.log,
                    "The number of derived parameters must be equal to the dimensionality"
                    " times the number of modes, i.e. %d x %d = %d, but was given %d "
                    "derived parameters.", self.d(), self.n_modes,
                    self.d() * self.n_modes,
                    len(self.output_params))
            elif not self.derived and self.output_params:
                raise LoggedError(
                    self.log,
                    "Derived parameters were requested, but 'derived' option is False. "
                    "Set to True and define as many derived parameters as the "
                    "dimensionality times the number of modes, i.e. %d x %d = %d.",
                    self.d(), self.n_modes, self.d() * self.n_modes)
        else:
            raise LoggedError(
                self.log, "You must specify both a mean (or a list of them) and a "
                          "covariance matrix, or a list of them.")
        self.gaussians = [multivariate_normal(mean=mean, cov=cov)
                          for mean, cov in zip(self.means, self.covs)]
        if self.weights:
            self.weights = np.asarray(self.weights)
            if not len(self.weights) == len(self.gaussians):
                raise LoggedError(self.log,
                                  "There must be as many weights as components.")
            if not np.isclose(sum(self.weights), 1):
                self.weights = self.weights / sum(self.weights)
                self.log.warning(
                    "Weights of components renormalized to %r", self.weights)
        else:
            self.weights = 1 / len(self.gaussians)

        # Prepare the transformation(s) for the derived parameters
        self.inv_choleskyL = [np.linalg.inv(np.linalg.cholesky(cov)) for cov in self.covs]

    def logp(self, **params_values):
        """
        Computes the log-likelihood for a given set of parameters.
        """
        self.wait()
        # Prepare the vector of sampled parameter values
        x = np.array([params_values[p] for p in self.input_params])
        # Fill the derived parameters
        derived = params_values.get("_derived")
        if derived is not None:
            n = self.d()
            for i in range(self.n_modes):
                standard = self.inv_choleskyL[i].dot(x - self.means[i])
                derived.update(
                    (p, v) for p, v in
                    zip(list(self.output_params)[i * n:(i + 1) * n], standard))
        # Compute the likelihood and return
        if len(self.gaussians) == 1:
            return self.gaussians[0].logpdf(x)
        else:
            return logsumexp([gauss.logpdf(x) for gauss in self.gaussians],
                             b=self.weights)


# Scripts to generate random means and covariances #######################################

def random_mean(ranges, n_modes=1, mpi_warn=True, random_state=None):
    """
    Returns a uniformly sampled point (as an array) within a list of bounds ``ranges``.

    The output of this function can be used directly as the value of the option ``mean``
    of the :class:`likelihoods.gaussian`.

    If ``n_modes>1``, returns an array of such points.
    """
    if not is_main_process() and mpi_warn:
        print("WARNING! "
              "Using with MPI: different process will produce different random results.")
    mean = np.array([uniform.rvs(loc=r[0], scale=r[1] - r[0], size=n_modes,
                                 random_state=random_state) for r in ranges])
    mean = mean.T
    if n_modes == 1:
        mean = mean[0]
    return mean


def random_cov(ranges, O_std_min=1e-2, O_std_max=1, n_modes=1,
               mpi_warn=True, random_state=None):
    """
    Returns a random covariance matrix, with standard deviations sampled log-uniformly
    from the length of the parameter ranges times ``O_std_min`` and ``O_std_max``, and
    uniformly sampled correlation coefficients between ``rho_min`` and ``rho_max``.

    The output of this function can be used directly as the value of the option ``cov`` of
    the :class:`likelihoods.gaussian`.

    If ``n_modes>1``, returns a list of such matrices.
    """
    if not is_main_process() and mpi_warn:
        print("WARNING! "
              "Using with MPI: different process will produce different random results.")
    dim = len(ranges)
    scales = np.array([r[1] - r[0] for r in ranges])
    cov = []
    for _ in range(n_modes):
        stds = scales * 10 ** (uniform.rvs(size=dim, loc=np.log10(O_std_min),
                                           scale=np.log10(O_std_max / O_std_min),
                                           random_state=random_state))
        this_cov = np.diag(stds).dot(
            (random_correlation.rvs(dim * stds / sum(stds), random_state=random_state)
             if dim > 1 else np.eye(1)).dot(np.diag(stds)))
        # Symmetrize (numerical noise is usually introduced in the last step)
        cov += [(this_cov + this_cov.T) / 2]
    if n_modes == 1:
        cov = cov[0]
    return cov


def info_random_gaussian_mixture(ranges, n_modes=1, input_params_prefix="",
                                 output_params_prefix="", O_std_min=1e-2, O_std_max=1,
                                 derived=False, mpi_aware=True,
                                 random_state=None, add_ref=False):
    """
    Wrapper around ``random_mean`` and ``random_cov`` to generate the likelihood and
    parameter info for a random Gaussian.

    If ``mpi_aware=True``, it draws the random stuff only once, and communicates it to
    the rest of the MPI processes.

    If ``add_ref=True`` (default: False) adds a reference pdf for the input parameters,
    provided that the gaussian mixture is unimodal (otherwise raises ``ValueError``).
    """
    cov: Any
    mean: Any
    if is_main_process() or not mpi_aware:
        cov = random_cov(ranges, n_modes=n_modes, O_std_min=O_std_min,
                         O_std_max=O_std_max, mpi_warn=False, random_state=random_state)
        if n_modes == 1:
            cov = [cov]
        # Make sure it stays away from the edges
        mean = [[]] * n_modes
        for i in range(n_modes):
            std = np.sqrt(cov[i].diagonal())
            factor = 3
            ranges_mean = [[r[0] + factor * s, r[1] - +factor * s] for r, s in
                           zip(ranges, std)]
            # If this implies min>max, take the centre
            ranges_mean = [
                (r if r[0] <= r[1] else 2 * [(r[0] + r[1]) / 2]) for r in ranges_mean]
            mean[i] = random_mean(ranges_mean, n_modes=1, mpi_warn=False,
                                  random_state=random_state)
    else:
        mean, cov = None, None
    if mpi_aware:
        mean, cov = share_mpi((mean, cov))
    dimension = len(ranges)
    info: InputDict = {"likelihood": {"gaussian_mixture": {
        "means": mean, "covs": cov, "input_params_prefix": input_params_prefix,
        "output_params_prefix": output_params_prefix, "derived": derived}},
        "params": dict(
            # sampled
            tuple((input_params_prefix + "_%d" % i,
                   {"prior": {"min": ranges[i][0], "max": ranges[i][1]},
                    "latex": r"\alpha_{%i}" % i})
                  for i in range(dimension)) +
            # derived
            (tuple((output_params_prefix + "_%d" % i,
                    {"latex": r"\beta_{%i}" % i})
                   for i in range(dimension * n_modes)) if derived else ()))}
    if add_ref:
        if n_modes > 1:
            raise ValueError("Cannot add a good reference pdf ('add_ref=True') for "
                             "multimodal distributions")
        for i, (p, v) in enumerate(info["params"].items()):
            v["ref"] = {"dist": "norm", "loc": mean[0][i], "scale": np.sqrt(cov[0][i, i])}
    return info



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path
from itertools import chain
import re
import sys

subfolders = {"likelihood": "likelihoods", "sampler": "samplers", "theory": "theories"}


def find_version():
    init_file = open(path.join(path.dirname(__file__), 'cobaya', '__init__.py')).read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", init_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def get_long_description():
    """Get the long description from the README file."""
    with open(path.join(path.abspath(path.dirname(__file__)), 'README.rst'),
              encoding='utf-8') as f:
        lines = f.readlines()
        i = -1
        while '=====' not in lines[i]:
            i -= 1
        return "".join(lines[:i])


def extract_docs_requirements():
    """Get requirements for building the documentation."""
    path_docs_requirements = path.join(path.abspath(path.dirname(__file__)),
                                       "docs/requirements.txt")
    # Quick fix for pip install from PyPI: apparently tries to run this function before
    # pulling the 'docs' folder. Assume this is only used with a git-pulled install.
    if not path.exists(path_docs_requirements):
        return []
    with open(path_docs_requirements, "r") as f:
        lines = f.readlines()
    i_first = next(i for i, line in enumerate(lines)
                   if line.strip().startswith("-e")) + 1
    reqs = [line.strip() for line in lines[i_first:]]
    return ["sphinx"] + reqs


install_requires = ['numpy>=1.17.0', 'scipy>=1.5', 'pandas>=1.0.1',
                    'PyYAML>=5.1', 'requests>=2.18', 'py-bobyqa>=1.2',
                    'GetDist>=1.3.1', 'fuzzywuzzy>=0.17', 'packaging', 'tqdm',
                    'portalocker>=2.3.0', 'dill>=0.3.3']

setup(
    name='cobaya',
    version=find_version(),
    description='Code for Bayesian Analysis',
    long_description=get_long_description(),
    url="https://cobaya.readthedocs.io",
    project_urls={
        'Source': 'https://github.com/CobayaSampler/cobaya',
        'Tracker': 'https://github.com/CobayaSampler/cobaya/issues',
        'Licensing': 'https://github.com/CobayaSampler/cobaya/blob/master/LICENCE.txt'
    },
    author="Jesus Torrado and Antony Lewis",
    license='LGPL',
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Astronomy',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11'
    ],
    python_requires='>=3.8.0',
    keywords='montecarlo sampling MCMC cosmology',
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=install_requires,
    extras_require={
        'test': ['pytest', 'pytest-forked', 'flaky', 'mpi4py'],
        'gui': ['pyside6', 'matplotlib'],
        'docs': extract_docs_requirements()},
    package_data={
        'cobaya': list(chain(*[['%s/*/*.yaml' % folder, '%s/*/*.bibtex' % folder]
                               for folder in subfolders.values()]))},
    entry_points={
        'console_scripts': [
            'cobaya-install=cobaya.install:install_script',
            'cobaya-create-image=cobaya.containers:create_image_script',
            'cobaya-prepare-data=cobaya.containers:prepare_data_script',
            'cobaya-run=cobaya.run:run_script',
            'cobaya-doc=cobaya.doc:doc_script',
            'cobaya-bib=cobaya.bib:bib_script',
            'cobaya-grid-create=cobaya.grid_tools:make_grid_script',
            'cobaya-grid-run=cobaya.grid_tools.runbatch:run',
            'cobaya-run-job=cobaya.grid_tools.runMPI:run_single',
            'cobaya-running-jobs=cobaya.grid_tools.running_jobs:running_jobs',
            'cobaya-delete-jobs=cobaya.grid_tools.delete_jobs:delete_jobs',
            'cobaya-cosmo-generator=cobaya.cosmo_input:gui_script',
        ],
    },
)



"""
.. module:: theory

:Synopsis: :class:`Theory` is a base class for theory codes and likelihoods.

Both likelihoods and theories calculate something. Likelihoods are distinguished
because they calculate a log likelihood. Both theory codes and likelihoods can calculate
other things, and they may have complex dependencies between them: e.g. the likelihood
depends on observable A that is computed by theory code B than in turn requires
calculation of input calculation by code C.

This module contains the base class for all of these calculation components. It handles
caching of results, so that calculations do not need to be redone when the parameters on
which a component directly (or indirectly) depends have not changed.

Subclasses generally provide the :meth:`Theory.get_requirements`,
:meth:`Theory.calculate` and initialization methods as required. The
:meth:`Theory.must_provide` method is used to tell a code which requirements are
actually needed by other components, and may return a dictionary of additional conditional
requirements based on those passed.

The :meth:`Theory.calculate` method saves all needed results in the state dictionary
(which is cached and reused as needed). Subclasses define ``get_X`` or ``get_result(X)``
methods to return the actual result of the calculation for X for the current cache state.
The :meth:`Theory.get_param` method returns the value of a derived parameter for the
current state.

For details and examples of how to handle multiple theory codes with complex dependencies
see :doc:`theories_and_dependencies`.

"""

from collections import deque
from typing import Sequence, Optional, Union, Tuple, Dict, Iterable, Set, Any, List
# Local
from cobaya.typing import TheoryDictIn, TheoriesDict, InfoDict, ParamValuesDict, \
    ParamsDict, empty_dict, unset_params
from cobaya.component import CobayaComponent, ComponentCollection, get_component_class
from cobaya.tools import str_to_list
from cobaya.log import LoggedError, always_stop_exceptions
from cobaya.tools import get_class_methods


class Theory(CobayaComponent):
    """Base class theory that can calculate something."""

    speed: float = -1
    stop_at_error: bool = False
    version: Optional[Union[dict, str]] = None
    params: ParamsDict

    # special components set by the dependency resolver;
    # (for Theory included in updated yaml but not in defaults)
    input_params: Sequence[str] = unset_params
    output_params: Sequence[str] = unset_params

    _states: deque

    def __init__(self, info: TheoryDictIn = empty_dict,
                 name: Optional[str] = None, timing: Optional[bool] = None,
                 packages_path: Optional[str] = None,
                 initialize=True, standalone=True):

        self._measured_speed = None
        super().__init__(info, name=name, timing=timing,
                         packages_path=packages_path, initialize=initialize,
                         standalone=standalone)

        # set to Provider instance before calculations
        self.provider: Any = None
        # Generate cache states, to avoid recomputing.
        # Default 3, but can be changed by sampler
        self.set_cache_size(3)
        self._helpers: Dict[str, 'Theory'] = {}
        self._input_params_extra: Set[str] = set()

    def get_requirements(self) -> Union[InfoDict, Sequence[str],
                                        Sequence[Tuple[str, InfoDict]]]:
        """
        Get a dictionary of requirements (or a list of requirement name, option tuples)
        that are always needed (e.g. must be calculated by another component
        or provided as input parameters).

        :return: dictionary or list of tuples of requirement names and options
                (or iterable of requirement names if no optional parameters are needed)
        """
        return str_to_list(getattr(self, "requires", []))

    def must_provide(self, **requirements) -> Union[None, InfoDict, Sequence[str],
                                                    Sequence[Tuple[str, InfoDict]]]:
        """
        Function called by Cobaya with the actual products that this component needs to
        compute (i.e. the things this component can provide that are actually used by
        other components). The function can return conditional requirements that this
        component needs from other components in order to compute those things.

        The ``requirements`` argument is a requirement name with any optional parameters.
        This function may be called more than once with different requirements.

        :return: optional dictionary (or list of requirement name, option tuples) of
                 conditional requirements for the ones requested.
        """
        # reset states whenever requirements change
        self._states.clear()
        return None

    def calculate(self, state, want_derived=True, **params_values_dict):
        """
        Do the actual calculation and store results in state dict

        :param state: dictionary to store results
        :param want_derived: whether to set state['derived'] derived parameters
        :param params_values_dict: parameter values
        :return: None or True if success, False for fail
        """

    def initialize_with_params(self):
        """
        Additional initialization after requirements called and input_params and
        output_params have been assigned (but provider and assigned requirements not yet
        set).
        """

    def initialize_with_provider(self, provider: 'Provider'):
        """
        Final initialization after parameters, provider and assigned requirements set.
        The provider is used to get the requirements of this theory using provider.get_X()
        and provider.get_param('Y').

        :param provider: the :class:`theory.Provider` instance that should be used by
                         this component to get computed requirements
        """
        self.provider = provider

    def get_param(self, p: str) -> float:
        """
        Interface function for likelihoods and other theory components to get derived
        parameters.
        """
        return self.current_state["derived"][p]

    def get_result(self, result_name, **kwargs):
        """
        Interface function for likelihood and other theory components to get
        quantities calculated by this component. By default assumes the quantity
        is just directly saved into the current state (i.e. returns
        ``state[result_name]``).

        :param result_name: name of quantity
        :param kwargs: options specific to X or this component
        :return: result
        """
        return self.current_state[result_name]

    def get_can_provide_methods(self):
        """
        Get a dictionary of quantities X that can be retrieved using get_X methods.

        :return: dictionary of the form {X: get_X method}
        """
        return get_class_methods(self.get_provider().__class__, not_base=Theory)

    def get_can_provide(self) -> Iterable[str]:
        """
        Get a list of names of quantities that can be retrieved using the general
        get_result(X) method.

        :return: iterable of quantity names
        """
        return []

    def get_can_provide_params(self) -> Iterable[str]:
        """
        Get a list of derived parameters that this component can calculate.
        The default implementation returns the result based on the params attribute set
        via the .yaml file or class params (with derived:True for derived parameters).

        :return: iterable of parameter names
        """
        params = getattr(self, "params", None)
        if params:
            return [k for k, v in params.items() if
                    hasattr(v, 'get') and v.get('derived') is True]
        else:
            return []

    def get_can_support_params(self) -> Iterable[str]:
        """
        Get a list of parameters supported by this component, can be used to support
        parameters that don't explicitly appear in the .yaml or class params attribute
        or are otherwise explicitly supported (e.g. via requirements)

        :return: iterable of names of parameters
        """
        return []

    def get_allow_agnostic(self):
        """
        Whether it is allowed to pass all unassigned input parameters to this
        component (True) or whether parameters must be explicitly specified (False).

        :return: True or False
        """
        return False

    @property
    def input_params_extra(self):
        """
        Parameters required from other components, to be passed as input parameters.
        """
        return self._input_params_extra

    def set_cache_size(self, n):
        """
        Set how many states to cache
        """
        self._states = deque(maxlen=n)

    def check_cache_and_compute(self, params_values_dict,
                                dependency_params=None, want_derived=False, cached=True):
        """
        Takes a dictionary of parameter values and computes the products needed by the
        likelihood, or uses the cached value if that exists for these parameters.
        If want_derived, the derived parameters are saved in the computed state
        (retrieved using current_derived).

        params_values_dict can be safely modified and stored.
        """

        if self._input_params_extra:
            params_values_dict.update(
                zip(self._input_params_extra,
                    self.provider.get_param(self._input_params_extra)))
        self.log.debug("Got parameters %r", params_values_dict)
        state = None
        if cached:
            for _state in self._states:
                if _state["params"] == params_values_dict and \
                        _state["dependency_params"] == dependency_params \
                        and (not want_derived or _state["derived"] is not None):
                    state = _state
                    self.log.debug("Re-using computed results")
                    self._states.remove(_state)
                    break
        if not state:
            self.log.debug("Computing new state")
            state = {"params": params_values_dict,
                     "dependency_params": dependency_params,
                     "derived": {} if want_derived else None}
            if self.timer:
                self.timer.start()
            try:
                if self.calculate(state, want_derived, **params_values_dict) is False:
                    return False
            except always_stop_exceptions:
                raise
            except Exception as excpt:
                if self.stop_at_error:
                    self.log.error("Error at evaluation. See error information below.")
                    raise
                else:
                    self.log.debug(
                        "Ignored error at evaluation and assigned 0 likelihood "
                        "(set 'stop_at_error: True' as an option for this component "
                        "to stop here and print a traceback). Error message: %r", excpt)
                    return False
            if self.timer:
                self.timer.increment(self.log)
        # make this state the current one
        self._states.appendleft(state)
        self._current_state = state
        return True

    @property
    def current_state(self) -> Dict:
        try:
            return self._current_state
        except AttributeError:
            raise LoggedError(self.log, "Cannot retrieve calculated quantities: "
                                        "nothing has been computed yet "
                                        "(maybe the prior was -infinity?)")

    @property
    def current_derived(self) -> ParamValuesDict:
        return self.current_state.get("derived", {})

    @property
    def type_list(self) -> List[str]:
        # list of labels that classify this component
        # not usually used for Theory, can be used for aggregated chi2 in likelihoods
        return str_to_list(getattr(self, "type", []) or [])

    def get_provider(self):
        """
        Return object containing get_X, get_param, get_result methods to get computed
        results.
        This defaults to self, but can change to delegate provision to another object

        :return: object instance
        """
        return self

    def get_helper_theories(self) -> Dict[str, 'Theory']:
        """
        Return dictionary of optional names and helper Theory instances that should be
        used in conjunction with this component. The helpers can be created here
        as only called once, and before any other use of helpers.

        :return: dictionary of names and Theory instances
        """
        return {}

    def update_for_helper_theories(self, helpers: Dict[str, 'Theory']):
        self._helpers = helpers
        if helpers:
            components: List[Theory] = list(helpers.values()) + [self]
            for output, attr in enumerate(("input_params", "output_params")):
                pars = getattr(self, attr, unset_params)
                if pars is not unset_params:
                    for component in components:
                        if not component.get_allow_agnostic():
                            if output:
                                supported = component.get_can_provide_params()
                            else:
                                supported = component.get_can_support_params()
                            setattr(component, attr, [p for p in pars if p in supported])
                            pars = [p for p in pars if p not in supported]
                    for component in components:
                        if component.get_allow_agnostic():
                            setattr(component, attr, pars)

    def get_attr_list_with_helpers(self, attr):
        """
        Get combined list of self.attr and helper.attr for all helper theories

        :param attr: attr name
        :return: combined list
        """
        values = list(getattr(self, attr))
        for helper in self._helpers.values():
            values.extend(getattr(helper, attr))
        return values

    def get_speed(self):
        return self._measured_speed or self.speed

    def set_measured_speed(self, speed):
        self.speed = speed


class TheoryCollection(ComponentCollection):
    """
    Initializes the list of theory codes.
    """

    def __init__(self, info_theory: TheoriesDict, packages_path=None, timing=None):
        super().__init__()
        self.set_logger("theory")

        if info_theory:
            for name, info in info_theory.items():
                info = info or {}
                # If it has an "external" key, wrap it up. Else, load it up
                if isinstance(info, Theory):
                    self.add_instance(name, info)
                elif isinstance(info.get("external"), Theory):
                    self.add_instance(name, info["external"])
                else:
                    if "external" in info:
                        theory_class = info["external"]
                        if not isinstance(theory_class, type) or \
                                not issubclass(theory_class, Theory):
                            raise LoggedError(self.log,
                                              "Theory %s is not a Theory subclass", name)
                    else:
                        theory_class = get_component_class(
                            name, kind="theory", class_name=info.get("class"),
                            logger=self.log)
                    self.add_instance(
                        name, theory_class(
                            info, packages_path=packages_path, timing=timing, name=name))

    def __getattribute__(self, name):
        if not name.startswith('_'):
            try:
                return super().__getattribute__(name)
            except AttributeError:
                self.log.warn("No attribute %s of TheoryCollection. Use model.provider "
                              "if you want to access computed requests" % name)
                pass
        return object.__getattribute__(self, name)


class HelperTheory(Theory):

    def has_version(self):
        # assume the main component handles all version checking
        return False


class Provider:
    """
    Class used to retrieve computed requirements.
    Just passes on get_X and get_param methods to the component assigned to compute them.

    For get_param it will also take values directly from the current sampling parameters
    if the parameter is defined there.
    """

    params: ParamValuesDict

    def __init__(self, model, requirement_providers: Dict[str, Theory]):
        self.model = model
        self.requirement_providers = requirement_providers
        self.params = {}

    def set_current_input_params(self, params):
        self.params = params

    def get_param(self, param: Union[str, Iterable[str]]) -> Union[float, List[float]]:
        """
        Returns the value of a derived (or sampled) parameter. If it is not a sampled
        parameter it calls :meth:`Theory.get_param` on component assigned to compute
        this derived parameter.

        :param param: parameter name, or a list of parameter names
        :return: value of parameter, or list of parameter values
        """
        if not isinstance(param, str):
            return [(self.params[p] if p in self.params else
                     self.requirement_providers[p].get_param(p)) for p in param]
        if param in self.params:
            return self.params[param]
        else:
            return self.requirement_providers[param].get_param(param)

    def get_result(self, result_name: str, **kwargs) -> Any:
        return self.requirement_providers[result_name].get_result(result_name, **kwargs)

    def __getattr__(self, name):
        if name.startswith('get_'):
            requirement = name[4:]
            try:
                return getattr(self.requirement_providers[requirement], name)
            except KeyError:  # requirement not listed (parameter or result)
                raise AttributeError
        return object.__getattribute__(self, name)
